#include "Camera.h"

#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>

Camera::Camera(const glm::vec3& position, const glm::vec3& target, const glm::vec3& up, int width, int height, float nearPlane, float farPlane, float fov)
	: _position(position), _target(target), _up(up),
	_width(width), _height(height), _ratio((float)width/(float)height),
	_nearPlane(nearPlane), _farPlane(farPlane),
	_fov(fov),
	_changedView(true), _changedPerspective(true)
{
	updateTransform();
}

void Camera::setPerspective(int width, int height, float nearPlane, float farPlane, float fov)
{
	_width=width;
	_height=height;
	_ratio = (float)width/(float)height;
	_nearPlane=nearPlane;
	_farPlane=farPlane;
	_fov=fov;
	_changedPerspective = true;
}

void Camera::setDistance(float nearPlane, float farPlane)
{
	_nearPlane=nearPlane;
	_farPlane=farPlane;
	_changedPerspective = true;
}

void Camera::setWindowSize(int width, int height)
{
	_width=width;
	_height=height;
	_ratio = (float)width/(float)height;
	_changedPerspective = true;
}

void Camera::setPosition(const glm::vec3& position)
{
	_position = position;
	_changedView = true;
}

void Camera::setTarget(const glm::vec3& target)
{
	_target = target;
	_changedView = true;
}

void Camera::setUp(const glm::vec3& up)
{
	_up = up;
	_changedView = true;
}

void Camera::setView(const glm::vec3& position, const glm::vec3& target, const glm::vec3& up)
{
	_position = position;
	_target = target;
	_up = up;
	_changedView = true;
}


// getter
const glm::vec3& Camera::getPosition() const
{
	return _position;
}

const glm::vec3& Camera::getTarget() const
{
	return _target;
}

const glm::vec3& Camera::getUp() const
{
	return _up;
}

int Camera::getWidth() const
{
	return _width;
}

int Camera::getHeight() const
{
	return _height;
}


const glm::mat4& Camera::getView() const
{
	return _view;
}

const glm::mat4& Camera::getInverseView() const
{
	return _invView;
}

const glm::mat4& Camera::getPerspective() const
{
	return _perspective;
}

const glm::mat4& Camera::getVP() const
{
	return _vp;
}


float Camera::getRatio() const
{
	return _ratio;
}

float Camera::getNearPlane() const
{
	return _nearPlane;
}

float Camera::getFarPlane() const
{
	return _farPlane;
}
float Camera::getFieldOfView() const
{
	return _fov;
}


void Camera::updateTransform()
{
	if (_changedView || _changedPerspective)
	{
		if (_changedView)
		{
			_view = glm::lookAt(_position, _target, _up);
			_invView = glm::inverse(_view);
			_changedView = false;
		}

		if (_changedPerspective)
		{
			_perspective = glm::perspective(_fov,_ratio,_nearPlane,_farPlane);
			_changedPerspective = false;
		}

		_vp = _perspective * _view;
	}
}

void Camera::getMouseRay(float mouseX, float mouseY, glm::vec3& rayOrigin, glm::vec3& rayDirection)
{
	float yHalfRange = tan(_fov*3.1415926535f/360.0f) * _nearPlane;
	float xHalfRange = yHalfRange * _ratio;

	glm::vec4 nearPlanePoint = glm::vec4(-xHalfRange, yHalfRange, -_nearPlane, 1.f);
	nearPlanePoint.x += (mouseX/_width) * 2.f * xHalfRange;
	nearPlanePoint.y -= (mouseY/_height) * 2.f * yHalfRange;

	rayOrigin = glm::vec3(_invView[3][0], _invView[3][1], _invView[3][2]);
	glm::vec3 wTarget = glm::vec3(_invView * nearPlanePoint);

	rayDirection = wTarget - rayOrigin;
}
