#ifndef _CAMERA_H_
#define _CAMERA_H_

#include <glm/glm.hpp>

/**
	Perspective projection camera.
	Encapsulates a position, direction, view frustum and screen size. Provides corresponding view and perspective matrices.
	The matrices are not automatically updated after a new value has been set. updateTransform() will update them.
*/
class Camera
{
public:
	/**
		Initializes position, direction, view frustum and screen size.
		@param fov opening angle of the camera in the vertical direction
	*/
	Camera(const glm::vec3& position, const glm::vec3& target, const glm::vec3& up, int width, int height, float nearPlane, float farPlane, float fov);

	// setter
	/**
		@param fov opening angle of the camera in the vertical direction
	*/
	void setPerspective(int width, int height, float nearPlane, float farPlane, float fov);
	/**
		sets near and far plane
	*/
	void setDistance(float nearPlane, float farPlane);
	void setWindowSize(int width, int height);
	void setPosition(const glm::vec3& position);
	void setTarget(const glm::vec3& target);
	void setUp(const glm::vec3& up);
	void setView(const glm::vec3& position, const glm::vec3& target, const glm::vec3& up);

	// getter
	const glm::vec3& getPosition() const;
	const glm::vec3& getTarget() const;
	const glm::vec3& getUp() const;

	const glm::mat4& getView() const;
	const glm::mat4& getInverseView() const;
	const glm::mat4& getPerspective() const;
	/**
		Returns the view projection matrix
		@return the view projection matrix
	*/
	const glm::mat4& getVP() const;

	/**
		Returns the screen's image ratio
		@return getWidth() / getHeight()
	*/
	float getRatio() const;
	float getNearPlane() const;
	float getFarPlane() const;
	/**
		Returns the opening angle of the camera in the vertical direction
		@return opening angle of the camera in the vertical direction
	*/
	float getFieldOfView() const;
	int getWidth() const;
	int getHeight() const;
	
	/**
		Returns the origin and direction of a ray shot from the camera position to the unprojection of the mouse position.
		The mouse coordinate system is thought to have (0,0) at the top-left corner of the frustum and (width,height) at the bottom-right corner.

		@param [in] mouseX x-coordinate of mouse position
		@param [in] mouseY y-coordinate of mouse position
		@param [out] rayOrigin origin of the mouse ray
		@param [out] rayDirection direction of the mouse ray
	*/
	void getMouseRay(float mouseX, float mouseY, glm::vec3& rayOrigin, glm::vec3& rayDirection);

	/**
		Recalculates the view matrix, perspective matrix and their inverses if any changes have been made to their parameters.
	*/
	void updateTransform();
private:
	int _width, _height; // window
	float _ratio;
	float _nearPlane, _farPlane;
	float _fov; // in degrees

	glm::vec3 _position;
	glm::vec3 _target;
	glm::vec3 _up;

	glm::mat4 _view; // look in direction Z-, Y+ is up
	glm::mat4 _invView;
	glm::mat4 _perspective;
	glm::mat4 _vp;

	bool _changedView, _changedPerspective;

};

#endif