#ifndef _CAMERACONTROLLER_H_
#define _CAMERACONTROLLER_H_

#include "Camera.h"
#include "InputHandler.h"
#include <glm/glm.hpp>


/**
	Interprets mouse movements to orbit, zoom and and change the target of a [Camera](@ref Camera) object.
	While the CameraController is in idle mode, the `update()` method will have no effect. While in orbit mode,
	the `update()` method will query the inputHandler for mouse movements and orbit around the selected
	pivot point accordingly.
*/

class CameraController
{
public:
	/**
		Initializes a CameraController object that will operate on the provided [Camera](@ref Camera) and will query the mouse position from the provided [InputHandler](@ref InputHandler).
	*/
	CameraController(Camera* camera, InputHandler* inputHandler);

	/**
		Activates orbit mode.
	*/
	void startOrbit();
	/**
		Goes to idle mode.
	*/
	void stopOrbit();
	/**
		Move the camera a little closer to the pivot if zoom is positive or away from the pivot if zoom is negative.
		@param zoom change in the mouse wheel position
	*/
	void zoomBy(int zoom);
	/**
		Should be called every frame.
		While the CameraController is in idle mode, the `update()` method will have no effect. While in orbit mode,
		the `update()` method will query the inputHandler for mouse movements and orbit around the selected
		pivot point accordingly.
		@param frameLength length of the last from in seconds.
	*/
	void update(double frameLength);
	/**
		Sets the point around which the camera orbits and zooms.
	*/
	void setPivot(const glm::vec3& pivot);

	/**
		Returns true iff not in idle mode.
	*/
	bool isActive() const;

private:

	enum CCState
	{
		IDLE,
		ORBIT
	};

	InputHandler* _inputHandler;
	Camera* _camera;
	glm::vec3 _pivot;
	glm::vec3 _position;

	glm::vec3 _startPosition;
	int _startMousePosX, _startMousePosY;
	int _lastMouseWheel;

	CCState _state;
};

#endif