#include "Commander.h"
#include "MeshLoader.h"
#include "SRTTransform.h"
#include "MeshInstance.h"
#include "StringUtil.h"
#include <fstream>
#include <iostream>
#include <algorithm>
#include <list>
#include <sstream>

Commander::Commander(Scene* scene)
	: _scene(scene)
{
}

Commander::~Commander()
{
}

Mesh* Commander::loadModel(const string& fileName)
{
	string lower = fileName;
	transform(fileName.begin(), fileName.end(), lower.begin(), ::tolower);

	Mesh* mesh = nullptr;
	unordered_map<string, Mesh*>::iterator iter = _meshes.find(lower);
	if (iter == _meshes.end())
	{
		mesh = new Mesh();

		MeshLoadDescription meshDescr;
		meshDescr.loadNormals = true;
		meshDescr.loadColors = true;
		meshDescr.moveOrigin = true;
		meshDescr.originX = meshDescr.originZ = Center;
		meshDescr.originY = Center;

		MeshLoader::loadCollada("models/"+fileName, mesh, meshDescr);
	
		_meshes[lower] = mesh;
		_meshNames[mesh] = lower;
	}
	else
	{
		mesh = iter->second;
	}

	return mesh;
}

SceneNode* Commander::addInstanceToScene(Mesh* mesh)
{
	SceneNode* node = _scene->createSceneNode();
	SRTTransform* transform = new SRTTransform(1, glm::vec3(0.f,0.f,0.f), glm::vec3(0.f,0.f,0.f));
	node->setTransform(transform);
	_scene->getRoot()->addSceneNode(node);

	MeshInstance* instance = new MeshInstance(mesh);
	node->addTransformable(instance);

	return node;
}

const string& Commander::getMeshName(Mesh* mesh)
{
	if (_meshNames.count(mesh) == 1)
	{
		return _meshNames[mesh];
	}
	return EMPTY_STRING;
}

void Commander::saveScene(const string& fileName)
{
	ofstream file("scenes/"+fileName, ios::trunc);

	if (file.is_open())
	{
		file << "global " << _scene->getGlobalAngle() << endl;
		file << endl;

		const list<SceneNode*>& children = _scene->getRoot()->getChildNodes();
		for (list<SceneNode*>::const_iterator iter = children.begin();
			iter != children.end();
			iter++)
		{
			SceneNode* child = *iter;
			const list<Transformable*>& transformables = child->getTransformables();
			if (transformables.size() > 0)
			{
				Transformable* x = *(transformables.begin());
				MeshInstance* mi = dynamic_cast<MeshInstance*>(x);
				if (mi)
				{
					Mesh* mesh = mi->getMesh();
					const string& meshName = getMeshName(mesh);
					SRTTransform* srt = (SRTTransform*) child->getTransform();
					glm::vec3& t = srt->t;
					float s = srt->s;
					glm::vec3& r = srt->r;

					file << "meshnode" << endl;
					file << "file " << meshName << endl;
					file << "t " << t.x << " " << t.y << " " << t.z << endl;
					file << "s " << s << endl;
					file << "r " << r.x << " " << r.y << " " << r.z << endl;
					file << "cancut " << (mi->canCut()?"on":"off") << endl;
					file << "canbecut " << (mi->canBeCut()?"on":"off") << endl;
					if (mi->getLocalAngle() >= 0.f)
					{
						file << "local " << mi->getLocalAngle() << endl;
					}
					file << "endmeshnode" << endl;
					file << endl;
				}
			}
		}
	}

	file.close();
}

void Commander::loadScene(const string& fileName)
{
	ifstream file("scenes/"+fileName);
	vector<string> tokens;
	istringstream iss;
	string line;
	glm::vec3 t,r;
	float s;
	float localAngle = -1.f;
	bool canCut, canBeCut;
	string modelfile;


	if (file.is_open())
	{
		_scene->clear();
		_scene->resetGlobalAngle();

		while (file.good())
		{
			getline(file,line);

			tokens.clear();
			StringUtil::tokenize(line, &tokens);

			if (tokens.size() == 0)
				continue;

			if (tokens[0] == "meshnode")
			{
				modelfile = "";
				t = glm::vec3(0.f);
				r = glm::vec3(0.f);
				s = 1.f;
				localAngle = -1.f;
				canCut = true;
				canBeCut = true;
			}
			else if (tokens[0] == "endmeshnode")
			{
				if (modelfile != "")
				{
					Mesh* mesh = loadModel(modelfile);
					SceneNode * node = addInstanceToScene(mesh);
					SRTTransform* srt = (SRTTransform*)node->getTransform();
					srt->t = t;
					srt->r = r;
					srt->s = s;
					if (node->getTransformables().size() > 0)
					{
						Transformable* transformable = *(node->getTransformables().begin());
						Renderable* renderable = dynamic_cast<Renderable*>(transformable);
						if (renderable)
						{
							renderable->setCanCut(canCut);
							renderable->setCanBeCut(canBeCut);
							if (localAngle >= 0.f)
							{
								renderable->setLocalAngle(localAngle);
							}
						}
					}
				}
			}
			else if (tokens.size() > 1)
			{
				if (tokens[0] == "file")
				{
					modelfile = tokens[1];
				}
				else if (tokens[0] == "t")
				{
					iss = istringstream(line.substr(2));
					iss >> t.x;
					iss >> t.y;
					iss >> t.z;
				}
				else if (tokens[0] == "r")
				{
					iss = istringstream(line.substr(2));
					iss >> r.x;
					iss >> r.y;
					iss >> r.z;
				}
				else if (tokens[0] == "s")
				{
					iss = istringstream(line.substr(2));
					iss >> s;
				}
				else if (tokens[0] == "cancut")
				{
					canCut = (tokens[1] != "off");
				}
				else if (tokens[0] == "canbecut")
				{
					canBeCut = (tokens[1] != "off");
				}
				else if (tokens[0] == "local")
				{
					iss = istringstream(tokens[1]);
					iss >> localAngle;
				}
				else if (tokens[0] == "global")
				{
					iss = istringstream(tokens[1]);
					float globalAngle;
					iss >> globalAngle;
					_scene->setGlobalAngle(globalAngle);
				}
			}
		}
	}

	file.close();
}

const string Commander::EMPTY_STRING = "";