#include "FontRenderer.h"
#include "Geometry.h"
#include <algorithm>
#include <iostream>

using namespace std;

bool compareGDDescendingArea(const GlyphDescription& gd1, const GlyphDescription& gd2)
{
	return gd1.width*gd1.height > gd2.width*gd2.height;
}

bool compareGDAscendingID(const GlyphDescription& gd1, const GlyphDescription& gd2)
{
	return gd1.glyphIndex < gd2.glyphIndex;
}

GlyphNode::GlyphNode()
	: gd(nullptr), left(0), top(0), width(0), height(0)
{
	child[0]=nullptr;
	child[1]=nullptr;
}

GlyphNode::~GlyphNode()
{
	if (child[0] != nullptr)
	{
		delete child[0];
		delete child[1];
	}
}

GlyphNode* GlyphNode::insert(GlyphDescription* gdInsert)
{
	// inner node
	if (child[0]!=nullptr)
	{
		GlyphNode* newNode = child[0]->insert(gdInsert);
		if (newNode != nullptr)
			return newNode;

		return child[1]->insert(gdInsert);
	}
	// leaf node
	else
	{
		// leaf occupied
		if (gd != nullptr)
			return nullptr;

		// glyph doesn't fit
		if (gdInsert->width > width || gdInsert->height > height)
			return nullptr;

		if (gdInsert->width == width && gdInsert->height == height)
		{
			gd = gdInsert;
			gdInsert->left = left;
			gdInsert->top = top;

			return this;
		}

		child[0] = new GlyphNode();
		child[1] = new GlyphNode();

		int dw = width - gdInsert->width;
		int dh = height - gdInsert->height;

		//  ______
		// | |    |
		// |_|____|
		//
		if (dw > dh)
		{
			child[0]->left = left;
			child[0]->width = gdInsert->width;
			child[0]->top = top;
			child[0]->height = height;

			child[1]->left = left+gdInsert->width;
			child[1]->width = dw;
			child[1]->top = top;
			child[1]->height = height;
		}
		//  _____
		// |_____|
		// |     |
		// |     |
		// |_____|
		//
		else
		{
			child[0]->left = left;
			child[0]->width = width;
			child[0]->top = top;
			child[0]->height = gdInsert->height;

			child[1]->left = left;
			child[1]->width = width;
			child[1]->top = top+gdInsert->height;
			child[1]->height = dh;
		}

		return child[0]->insert(gdInsert);
	}
}

FontRenderer::FontRenderer()
	: fontShader("Font Shader")
{
	if(FT_Init_FreeType(&_ftLibrary))
	{
		cout << "ERR: Could not init FreeType library" << endl;
		return;
	}

	glPrimitiveRestartIndex(-1);
	glEnable(GL_PRIMITIVE_RESTART);

	fontShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/fontshader.vert");
	fontShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/fontshader.frag");
	fontShader.compileShaders();
	fontShader.link("outColor");
}

void FontRenderer::loadFont(const std::string& fontPath, const std::string& fontName, int pixelHeight)
{
	FT_Face face;
	if (FT_New_Face(_ftLibrary, fontPath.c_str(), 0, &face))
	{
		cout << "ERR: Could not load font face" << endl;
		return;
	}

	if (FT_Set_Pixel_Sizes(face, 0, pixelHeight))
	{
		cout << "ERR: Could not load font with height of " << pixelHeight << "px" << endl;
		return;
	}

	FT_GlyphSlot g = face->glyph;
	std::vector<GlyphDescription>* glyphs = new std::vector<GlyphDescription>();

	// Load metrics of all glyphs
	int totalArea = 0;
	for(FT_UInt i=0; i < (FT_UInt)face->num_glyphs; i++)
	{
		if (!FT_Load_Glyph(face, i, FT_LOAD_DEFAULT))
		{
			GlyphDescription gd;
			gd.glyphIndex = i;
			gd.width = g->metrics.width/64;
			gd.height = g->metrics.height/64;
			gd.bearingX = g->metrics.horiBearingX/64;
			gd.bearingY = g->metrics.horiBearingY/64;
			gd.advance = g->metrics.horiAdvance/64;
			gd.left = 0;
			gd.top = 0;

			glyphs->push_back(gd);

			totalArea += gd.width*gd.height;
		}
	}

	std::sort(glyphs->begin(), glyphs->end(), compareGDDescendingArea);

	// Pack glyphs into kdtree
	int insertedGlyphs=0;
	int maxSide=0;
	float multiplier = 1.05f;
	do {
		insertedGlyphs = 0;
		maxSide = (int)sqrt(totalArea*multiplier);
		multiplier += 0.05f;

		GlyphNode* rootNode = new GlyphNode();
		rootNode->width = maxSide;
		rootNode->height = maxSide;

		for(vector<GlyphDescription>::iterator iter = glyphs->begin();
			iter != glyphs->end();
			iter++)
		{
			insertedGlyphs += rootNode->insert(&*iter)==nullptr?0:1;
		}

		delete rootNode;
	} while (insertedGlyphs != glyphs->size());

	std::sort(glyphs->begin(), glyphs->end(), compareGDAscendingID);

	// dump glyphs into atlas
	unsigned char *atlas = new unsigned char[maxSide*maxSide];
	std::fill_n(atlas,maxSide*maxSide,0);

	for(vector<GlyphDescription>::iterator iter = glyphs->begin();
		iter != glyphs->end();
		iter++)
	{
		int glyphIndex = iter->glyphIndex;
		int left = iter->left;
		int top = iter->top;
		int width = iter->width;
		int height = iter->height;

		if (!FT_Load_Glyph(face, glyphIndex, FT_LOAD_RENDER))
		{
			for (int i=0;i<height;i++)
			{
				for (int j=0;j<width;j++)
				{
					atlas[maxSide*(top+i)+left+j] = g->bitmap.buffer[width*i+j];
				}
			}
		}
	}


	// put atlas into texture
	GLuint texHandle;
	glGenTextures(1, &texHandle);
	glBindTexture(GL_TEXTURE_2D, texHandle);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	
	glTexImage2D(
		GL_TEXTURE_2D,
		0,
		GL_RED,
		maxSide,
		maxSide,
		0,
		GL_RED,
		GL_UNSIGNED_BYTE,
		atlas
	);

	glBindTexture(GL_TEXTURE_2D, 0);

	delete[] atlas;

	FaceDescription fd;
	fd.fontPath = fontPath;
	fd.pixelHeight = pixelHeight;
	fd.glyphs = glyphs;
	fd.textureHandle = texHandle;
	fd.textureWidth = fd.textureHeight = maxSide;
	fd.face = face;

	_faces[fontName] = fd;
}

void FontRenderer::render(Shader* shader)
{
	FaceDescription& fd = _faces["Lucida"];

	Mesh* mesh = new Mesh();
	Geometry::createScreenQuad(mesh, true);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, fd.textureHandle);
	glUniform1i(shader->getUniformLocation("texture0"), 0);

	mesh->bindVAO();
	mesh->draw();
	mesh->unbindVAO();

	delete mesh;
}

void FontRenderer::createText(const std::string& text, const std::string& fontName, Scene* scene, Text* textObj)
{
	textObj->init(text, &(_faces[fontName]), scene, &fontShader);
}

Text::Text()
	: _text(""), _faceDescr(nullptr), _mesh(nullptr), _scene(nullptr)
{
}

Text::~Text()
{
	if (_mesh != nullptr)
		delete _mesh;
}

void Text::init(const std::string& text, FaceDescription* faceDescr, Scene* scene, Shader* shader)
{
	_faceDescr = faceDescr;
	_scene = scene;
	_shader = shader;

	setText(text);
}

void Text::setText(const std::string& text)
{
	_text = text;

	float invScreenWidth = 1.f;//1.f / (float)_scene->getCamera()->getWidth();
	float invScreenHeight = 1.f;//1.f / (float)_scene->getCamera()->getHeight();
	float invTexWidth = 1.f / (float)_faceDescr->textureWidth;
	float invTexHeight = 1.f / (float)_faceDescr->textureHeight;

	float penX = 0.f, penY = 0.f;

	GLfloat *vertices = new GLfloat[12 * _text.length()];
	GLfloat *texCoords = new GLfloat[8 * _text.length()];
	GLuint *indices = new GLuint[7 * _text.length()];

	int ti = 0, vi = 0, ii = 0;
	for(int i=0; i < (int)_text.length(); i++)
	{
		FT_UInt glyph = FT_Get_Char_Index(_faceDescr->face, _text[i]);
		GlyphDescription& gd = _faceDescr->glyphs->at(glyph);

		float leftT = (float)gd.left * invTexWidth;
		float rightT = (float)(gd.left+gd.width) * invTexWidth;
		float bottomT = (float)(gd.top+gd.height) * invTexHeight;
		float topT = (float)gd.top * invTexHeight;

		float leftV = penX + (float)gd.bearingX * invScreenWidth;
		float rightV = leftV + (float)gd.width * invScreenWidth;
		float topV = penY + (float)gd.bearingY * invScreenHeight;
		float bottomV = topV - (float)gd.height * invScreenHeight;
		
		texCoords[ti++] = leftT;
		texCoords[ti++] = topT;
		texCoords[ti++] = leftT;
		texCoords[ti++] = bottomT;
		texCoords[ti++] = rightT;
		texCoords[ti++] = topT;
		texCoords[ti++] = rightT;
		texCoords[ti++] = bottomT;

		vertices[vi++] = leftV;
		vertices[vi++] = topV;
		vertices[vi++] = 0.f;
		vertices[vi++] = leftV;
		vertices[vi++] = bottomV;
		vertices[vi++] = 0.f;
		vertices[vi++] = rightV;
		vertices[vi++] = topV;
		vertices[vi++] = 0.f;
		vertices[vi++] = rightV;
		vertices[vi++] = bottomV;
		vertices[vi++] = 0.f;

		int fi = i * 4;

		indices[ii++] = fi;
		indices[ii++] = fi+1;
		indices[ii++] = fi+2;
		indices[ii++] = fi+2;
		indices[ii++] = fi+1;
		indices[ii++] = fi+3;
		indices[ii++] = -1;

		penX += gd.advance * invScreenWidth;
	}

	if (_mesh != nullptr)
		delete _mesh;
	_mesh = new Mesh();
	_mesh->init(_text.length()*4, vertices, nullptr, texCoords, _text.length()*7, indices, GL_TRIANGLES);

	delete[] vertices;
	delete[] texCoords;
	delete[] indices;
}

const std::string& Text::getText() const
{
	return _text;
}

void Text::render()
{
	_shader->bind();

	glUniform2f(_shader->getUniformLocation("invImageSize"), 1.f/_scene->getCamera()->getWidth(), 1.f/_scene->getCamera()->getHeight());
	glUniform2f(_shader->getUniformLocation("pOrigin"), (float)_pX, (float)_pY);
	glUniform3f(_shader->getUniformLocation("color"), 0.f,0.f,0.f);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, _faceDescr->textureHandle);
	glUniform1i(_shader->getUniformLocation("texture0"), 0);

	_mesh->bindVAO();
	_mesh->draw();
	_mesh->unbindVAO();

	_shader->unbind();
}

void Text::setPosition(int pX, int pY)
{
	_pX = pX;
	_pY = pY;
}