#include "Framebuffer.h"

#include <iostream>

using namespace std;

FramebufferDescriptor::FramebufferDescriptor(int width, int height)
	: width(width), height(height), depthFormat(GL_DEPTH_COMPONENT), depthInternalFormat(GL_DEPTH_COMPONENT16),
	numColorAttachments(1), colorFormat(nullptr), colorInternalFormat(nullptr), textureWrap(GL_CLAMP_TO_EDGE),
	depthMinFilter(GL_NEAREST), depthMagFilter(GL_NEAREST), depthType(GL_UNSIGNED_INT),
	colorMinFilter(GL_NEAREST), colorMagFilter(GL_NEAREST), colorType(GL_UNSIGNED_BYTE)
{
}

Framebuffer::Framebuffer()
		: _width(-1), _height(-1), _depthFormat(-1), _depthInternalFormat(-1),
		_numColorAttachments(-1), _colorFormat(nullptr), _colorInternalFormat(nullptr),
		_framebufferHandle(-1), _depthTextureHandle(-1), _colorTextureHandle(nullptr),
		_initted(false), _colorAttachments(nullptr)
{
}

bool Framebuffer::init(const FramebufferDescriptor& descriptor)
{
	if (_initted || descriptor.width == 0 || descriptor.height == 0)
		return false;

	_width = descriptor.width;
	_height = descriptor.height;
	_numColorAttachments = descriptor.numColorAttachments;

	if (_numColorAttachments > 0)
	{
		_colorFormat = new int[_numColorAttachments];
		_colorInternalFormat = new int[_numColorAttachments];
		_colorTextureHandle = new GLuint[_numColorAttachments];
	}

	_initted = true;

	for (int i=0; i<descriptor.numColorAttachments; i++)
	{
		_colorFormat[i] = descriptor.colorFormat[i];
		_colorInternalFormat[i] = descriptor.colorInternalFormat[i];
	}

	// create framebuffer
	glGenFramebuffers(1, &_framebufferHandle);
	glBindFramebuffer(GL_FRAMEBUFFER, _framebufferHandle);

	// depth attachment
	_depthFormat = descriptor.depthFormat;
	_depthInternalFormat = descriptor.depthInternalFormat;

	glGenTextures(1, &_depthTextureHandle);
	glBindTexture(GL_TEXTURE_2D, _depthTextureHandle);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, descriptor.depthMinFilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, descriptor.depthMagFilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, descriptor.textureWrap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, descriptor.textureWrap);

	glTexImage2D(GL_TEXTURE_2D, 0, _depthInternalFormat, _width, _height, 0, _depthFormat, descriptor.depthType, NULL);

	if (_depthFormat == GL_DEPTH_STENCIL)
	{
		glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, _depthTextureHandle, 0);
	}
	else
	{
		glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, _depthTextureHandle, 0);
	}

	// color attachments
	if (_numColorAttachments > 0)
		glGenTextures(_numColorAttachments, _colorTextureHandle);

	for (int i=0; i<_numColorAttachments; i++)
	{
		glBindTexture(GL_TEXTURE_2D, _colorTextureHandle[i]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, descriptor.colorMinFilter);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, descriptor.colorMagFilter);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, descriptor.textureWrap);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, descriptor.textureWrap);
		glTexImage2D(GL_TEXTURE_2D, 0, _colorInternalFormat[i], _width, _height, 0, _colorFormat[i], descriptor.colorType, NULL);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+i, _colorTextureHandle[i], 0);
	}
	if (_numColorAttachments > 0)
	{
		_colorAttachments = new GLenum[_numColorAttachments];
		for (int i=0; i<_numColorAttachments; i++)
		{
			_colorAttachments[i] = GL_COLOR_ATTACHMENT0 + i;
		}
	}
	glBindTexture(GL_TEXTURE_2D, 0);

	bool ok = false;

	GLuint status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	switch(status) {
	case GL_FRAMEBUFFER_COMPLETE:
		ok = true;
		break;

	case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
		cout << "An attachment could not be bound to frame buffer object!" << endl;
		break;

	case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
		cout <<  "Attachments are missing! At least one image (texture) must be bound to the frame buffer object!" << endl;
		break;

	/*case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
		cout << "The dimensions of the buffers attached to the currently used frame buffer object do not match!" << endl;
		break;

	case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
		cout << "The formats of the currently used frame buffer object are not supported or do not fit together!" << endl;
		break;*/

	case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER:
		cout << "A Draw buffer is incomplete or undefinied. All draw buffers must specify attachment points that have images attached." << endl;
		break;

	case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER:
		cout << "A Read buffer is incomplete or undefinied. All read buffers must specify attachment points that have images attached." << endl;
		break;

	case GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE:
		cout << "All images must have the same number of multisample samples." << endl;
		break;

	case GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS :
		cout << "If a layered image is attached to one attachment, then all attachments must be layered attachments. The attached layers do not have to have the same number of layers, nor do the layers have to come from the same kind of texture." << endl;
		break;

	case GL_FRAMEBUFFER_UNSUPPORTED:
		cout << "Attempt to use an unsupported format combinaton!" << endl;
		break;

	default:
		cout << "Unknown error while attempting to create frame buffer object!" << endl;
		break;
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	return ok;
}

Framebuffer::~Framebuffer()
{
	if (_initted)
	{
		if (_numColorAttachments > 0)
		{
			delete[] _colorFormat;
			delete[] _colorInternalFormat;
			delete[] _colorTextureHandle;
			delete[] _colorAttachments;
			glDeleteTextures(_numColorAttachments, _colorTextureHandle);
		}
		glDeleteTextures(1, &_depthTextureHandle);
		glDeleteFramebuffers(1, &_framebufferHandle);
	}
}

void Framebuffer::bind() const
{
	glBindFramebuffer(GL_FRAMEBUFFER, _framebufferHandle);
	if (_numColorAttachments == 0)
	{
		glDrawBuffer(GL_NONE);
	} else {
		glDrawBuffers(_numColorAttachments, _colorAttachments);
	}
}

void Framebuffer::unbind() const
{
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glDrawBuffer(GL_BACK);
}

GLuint Framebuffer::getDepthTextureHandle() const
{
	return _depthTextureHandle;
}

GLuint Framebuffer::getColorTextureHandle(int i) const
{
	return _colorTextureHandle[i];
}

GLuint Framebuffer::getHandle() const
{
	return _framebufferHandle;
}

void Framebuffer::setViewport() const
{
	glViewport(0,0,_width,_height);
}

int Framebuffer::getWidth() const
{
	return _width;
}

int Framebuffer::getHeight() const
{
	return _height;
}