#ifndef _GLFWGAMEWINDOW_H_
#define _GLFWGAMEWINDOW_H_

#include <string>

/**
	Base class for a GLFW window with OpenGL core context 3.0+ and a generic render loop.
*/
class GLFWGameWindow
{
public:
	/**
		Initializes fields of the objects. Call `run()` to start the render loop.
		@param width window width in pixels
		@param height window height in pixels
		@param title window title
		@param fullscreen false for windowed mode, true for fullscreen mode
		@param major major version of the GL context, 3.x or 4.x
		@param minor minor version of the GL context
	*/
	GLFWGameWindow(int width, int height, const std::string& title, bool fullscreen, int major, int minor);
	~GLFWGameWindow();

	/**
		Purely virtual function, to initialize resources after OpenGL context has been creates
	*/
	virtual void init() = 0;
	/**
		Purely virtual function, to perform the application logic each frame
	*/
	virtual void update(double frameLength) = 0;
	/**
		Purely virtual function, to render each frame
	*/
	virtual void draw() = 0;
	/**
		Purely virtual function, to deallocate resources before window destruction
	*/
	virtual void destroy() = 0;

	/**
		Creates an OpenGL context, calls `init` and starts the render loop.
	*/
	bool run();

	/**
		Causes the render loop to terminate after its current iteration.
	*/
	void stop();

	void setGLVersion(int major, int minor);
	/**
		if set to true, the current fps will be displayed to the standard output every few seconds
	*/
	void setPrintFPS(bool print);

	double getAverageFPS() const;

	int getWidth() const;
	int getHeight() const;

private:
	int _width, _height;
	bool _fullscreen;
	std::string _title;
	double _maxFrameLength;
	double _minFrameLength;
	double _invFPS;
	bool _running;

	int _glMajor, _glMinor;
	bool _printFPS;
	double _printFPSInterval;
};

#endif