#include "Gizmo.h"
#include "Camera.h"
#include "Scene.h"
#include "SRTTransform.h"
#include <iostream>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <gl/glfw.h>

using namespace std;

Gizmo::Gizmo(Scene* scene)
	: _visible(false), _scene(scene)
{
	_tX = new TranslateGizmo(scene,0);
	_tY = new TranslateGizmo(scene,1);
	_tZ = new TranslateGizmo(scene,2);
	_s = new ScaleGizmo(scene);
}

void Gizmo::setVisible(bool visible)
{
	_visible = visible;
}

bool Gizmo::isVisible()
{
	return _visible;
}

void Gizmo::setTarget(SceneNode* target)
{
	_target = target;
	_tX->setTarget(target);
	_tY->setTarget(target);
	_tZ->setTarget(target);
	_s->setTarget(target);
}

void Gizmo::update()
{
	_tX->update();
	_tY->update();
	_tZ->update();
	_s->update();
}

void Gizmo::render(Shader* shader)
{
	if (_visible && _target != nullptr)
	{
		glm::mat4 transform = _target->getTotalTransform();
		float dist = sqrt(transform[3][0]*transform[3][0]
			+transform[3][1]*transform[3][1]
			+transform[3][2]*transform[3][2]);
		float scale = _target->getTotalScale();

		for (int i=0;i<3;i++)
			for (int j=0;j<3;j++)
				transform[i][j] *= 0.05f*dist/scale;
		const glm::mat4& viewMatrix = _scene->getCamera()->getView();
		glm::vec3 cam = glm::vec3(viewMatrix[3][0],viewMatrix[3][1],viewMatrix[3][2]);

		glUniformMatrix4fv(shader->getUniformLocation("modelView"), 1, GL_FALSE, glm::value_ptr(transform));
		glUniformMatrix3fv(shader->getUniformLocation("normalMatrix"), 1, GL_FALSE, glm::value_ptr(glm::mat3(transform)));

		_tX->render(shader);
		_tY->render(shader);
		_tZ->render(shader);
		_s->render(shader);
	}
}

void Gizmo::stopEdit()
{
	_tX->stopEdit();
	_tY->stopEdit();
	_tZ->stopEdit();
	_s->stopEdit();
}

SubGizmo::SubGizmo(Scene* scene)
	: _scene(scene), _state(IDLE)
{
}

void SubGizmo::setTarget(SceneNode* target)
{
	_target = target;
}

SceneNode* SubGizmo::getTarget()
{
	return _target;
}

Scene* SubGizmo::getScene()
{
	return _scene;
}


void SubGizmo::hover()
{
	if (!isEditing())
		_state = HOVERING;
}

void SubGizmo::stopEdit()
{
	setIdle();
}

bool SubGizmo::isIdle()
{
	return _state==IDLE;
}

bool SubGizmo::isHovering()
{
	return _state==HOVERING;
}

bool SubGizmo::isEditing()
{
	return _state==EDITING;
}

void SubGizmo::setIdle()
{
	_state=IDLE;
}

void SubGizmo::setEditing()
{
	_state=EDITING;
}

ScaleGizmo::ScaleGizmo(Scene* scene)
	: SubGizmo(scene), _mesh(new Mesh())
{
	MeshDescriptor outDescr;
	outDescr.numVertices = 8;
	outDescr.vertices = VERTEX_POSITION;
	outDescr.normals = nullptr;
	outDescr.colors = VERTEX_COLOR;
	outDescr.texCoords = nullptr;
	outDescr.numIndices = 17;
	outDescr.indices = VERTEX_INDICES;
	outDescr.hasMinMax = false;
	outDescr.primitiveMode = GL_TRIANGLE_STRIP;

	_mesh->init(outDescr);
}

void ScaleGizmo::render(Shader* shader)
{
	glUniform1ui(shader->getUniformLocation("colorID"), getID());
	glUniform3f(shader->getUniformLocation("color"), 1.f,1.f,1.f);
	_mesh->bindVAO();
	_mesh->draw();
	_mesh->unbindVAO();
}

UserAction ScaleGizmo::press()
{
	setEditing();

	glfwGetMousePos(&_mouseStartX, &_mouseStartY);
	SRTTransform* srt = (SRTTransform*) getTarget()->getTransform();
	_scaleStart = srt->s;

	return GIZMO_EDIT;
}

void ScaleGizmo::update()
{
	glm::vec3 color(1.f,1.f,1.f);
	if (isHovering())
	{
		color *= 1.4f;
		setIdle();
	}
	else if (isEditing())
	{
		color *= 1.4f;

		int xPos, yPos;
		glfwGetMousePos(&xPos, &yPos);

		int dx = xPos-_mouseStartX;
		float scaleFactor = 1.f;
		if (dx > 0)
		{
			scaleFactor = (dx+100)/100.f;
		}
		else
		{
			scaleFactor = 100.f/(100-dx);
		}

		if (getTarget()->getParent() != nullptr)
		{
			SRTTransform* srt = (SRTTransform*) getTarget()->getTransform();
			srt->s = _scaleStart * scaleFactor;
			srt->update();
		}
	}
}

TranslateGizmo::TranslateGizmo(Scene* scene, int axis)
	: SubGizmo(scene), _mesh(new Mesh()), _axis(axis),
	_translationPlane(glm::vec3(0.f,0.f,0.f), glm::vec3(1.f,0.f,0.f))
{
	// init mesh

	MeshDescriptor outDescr;
	outDescr.numVertices = 4;
	outDescr.normals = nullptr;
	outDescr.colors = VERTEX_COLOR;
	outDescr.texCoords = nullptr;
	outDescr.indices = nullptr;
	outDescr.hasMinMax = false;
	outDescr.primitiveMode = GL_TRIANGLE_STRIP;

	if (axis==0)
	{
		outDescr.vertices = VERTEX_POSITION_X;
		_color = glm::vec3(1.f,0.5f,0.5f);
	}
	else if (axis==1)
	{
		outDescr.vertices = VERTEX_POSITION_Y;
		_color = glm::vec3(0.5f,1.0f,0.5f);
	}
	else if (axis==2)
	{
		outDescr.vertices = VERTEX_POSITION_Z;
		_color = glm::vec3(0.5f,0.5f,1.0f);
	}

	_mesh->init(outDescr);
}

void TranslateGizmo::update()
{
	_color = glm::vec3(0.4f);
	_color[_axis] = 0.8f;
	if (isHovering())
	{
		_color *= 1.4f;
		setIdle();
	}
	else if (isEditing())
	{
		_color *= 1.4f;

		int xPos, yPos;
		glfwGetMousePos(&xPos, &yPos);

		float mTargetX = (float)xPos + _mouseOffsetX;
		float mTargetY = (float)yPos + _mouseOffsetY;

		glm::vec3 rayOrigin, rayDir;
		getScene()->getCamera()->getMouseRay(mTargetX, mTargetY, rayOrigin, rayDir);

		glm::vec3 intersection;
		_translationPlane.lineIntersect(rayOrigin, rayDir, intersection);

		if (getTarget()->getParent() != nullptr)
		{
			glm::vec4 modelTranslate = getScene()->getCamera()->getView() * glm::vec4(intersection, 1.f);
			modelTranslate = glm::inverse(getTarget()->getParent()->getTotalTransform()) * modelTranslate;
			SRTTransform* srt = (SRTTransform*) getTarget()->getTransform();
			srt->t = glm::vec3(modelTranslate);
			srt->update();
		}
	}
}

void TranslateGizmo::render(Shader* shader)
{
	glUniform1ui(shader->getUniformLocation("colorID"), getID());
	glUniform3f(shader->getUniformLocation("color"), _color.x, _color.y, _color.z);
	_mesh->bindVAO();
	_mesh->draw();
	_mesh->unbindVAO();
}

UserAction TranslateGizmo::press()
{
	//_scene->startGizmoEdit();
	setEditing();

	int xPos, yPos;
	glfwGetMousePos(&xPos, &yPos);

	Camera* cam = getScene()->getCamera();

	const glm::mat4& mv = getTarget()->getTotalTransform();
	glm::vec4 vOrigin = glm::vec4(mv[3][0],mv[3][1],mv[3][2],1.f);
	glm::vec4 pOrigin = cam->getPerspective() * vOrigin;
	pOrigin /= pOrigin.w;

	glm::vec2 mOrigin = glm::vec2((0.5f+0.5f*pOrigin.x)*cam->getWidth(), (0.5f-0.5f*pOrigin.y)*cam->getHeight());
	
	_mouseOffsetX = mOrigin.x - (float)xPos;
	_mouseOffsetY = mOrigin.y - (float)yPos;

	glm::vec3 planeOrigin = glm::vec3(cam->getInverseView() * vOrigin);
	glm::vec3 planeNormal = glm::vec3(0.f,0.f,0.f);
	planeNormal[_axis] = 1.f;

	_translationPlane = Plane(planeOrigin, planeNormal);

	return GIZMO_EDIT;
}

GLfloat TranslateGizmo::VERTEX_POSITION_X[] = {
	.0f, -1.f,  1.f,
	.0f, -1.f, -1.f,
	.0f,  1.f,  1.f,
	.0f,  1.f, -1.f,
};

GLfloat TranslateGizmo::VERTEX_POSITION_Y[] = {
	-1.f, .0f,  1.f,
	-1.f, .0f, -1.f,
	 1.f, .0f,  1.f,
	 1.f, .0f, -1.f,
};

GLfloat TranslateGizmo::VERTEX_POSITION_Z[] = {
	-1.f,  1.f, .0f,
	-1.f, -1.f, .0f,
	 1.f,  1.f, .0f,
	 1.f, -1.f, .0f
};

GLfloat TranslateGizmo::VERTEX_COLOR[] = {
	1.f,1.f,1.f,
	1.f,1.f,1.f,
	1.f,1.f,1.f,
	1.f,1.f,1.f
};

GLfloat ScaleGizmo::VERTEX_POSITION[] = {
	-0.4f,-0.4f,0.4f,
	0.4f,-0.4f,0.4f,
	-0.4f,0.4f,0.4f,
	0.4f,0.4f,0.4f,
	-0.4f,0.4f,-0.4f,
	0.4f,0.4f,-0.4f,
	-0.4f,-0.4f,-0.4f,
	0.4f,-0.4f,-0.4f
};

GLfloat ScaleGizmo::VERTEX_COLOR[] = {
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f,
	0.7f,0.7f,0.7f
};

GLuint ScaleGizmo::VERTEX_INDICES[] = {
	0,1,2,3,4,5,6,7,
	-1,
	5,3,7,1,6,0,4,2
};