#ifndef _GIZMO_H_
#define _GIZMO_H_

#include "Identifiable.h"
#include "Renderable.h"
#include "Mesh.h"
#include "SceneNode.h"
#include "Plane.h"
#include <glm/glm.hpp>

class TranslateGizmo;
class ScaleGizmo;
class Scene;

/**
	Draws a handle on top of an object that allows the user to manipulate its position and scale.

	@see SubGizmo
*/
class Gizmo : public Renderable
{
public:
	Gizmo(Scene* scene);

	/**
		Sets the scene node, whose properties the user wants to edit.
	*/
	void setTarget(SceneNode* target);
	/**
		Renders its subcomponents
		@see SubGizmo
	*/
	void render(Shader* shader);
	/**
		Should be called every frame
	*/
	void update();

	void setVisible(bool visible);
	bool isVisible();

	/**
		Finished any editing operation in process
	*/
	void stopEdit();

private:
	SceneNode* _target;
	TranslateGizmo *_tX, *_tY, *_tZ;
	ScaleGizmo* _s;
	bool _visible;

	Scene* _scene;
};

/**
	Base class for gizmos that allow editing in a specific way.
*/
class SubGizmo : public Identifiable
{
public:
	friend class Gizmo;

	SubGizmo(Scene* scene);

	Scene* getScene();

	/**
		Renders a representation of the gizmo ontop of its target
	*/
	virtual void render(Shader* shader)=0;
	/**
		Called by the application before the `update()` method iff the mouse cursor is resting on the sub gizmo
	*/
	virtual void hover();
	/**
		Called by the application when the user clicks on the gizmo. Allows to grab the focus of the application
	*/
	virtual UserAction press()=0;

	/**
		Called be the surrounding `Gizmo` to finish the editing process.
	*/
	void stopEdit();
	/**
		Called be the surrounding `Gizmo` to set the scene node that will be edited.
	*/
	void setTarget(SceneNode* target);
	/**
		Returns the scene node that can be edited by this gizmo.
	*/
	SceneNode* getTarget();

	/**
		Has to be implemented by a sub class to perform its logic during the editing process
	*/
	virtual void update()=0;

	/**
		returns true iff no editing operation is in process and the mouse cursor is not resting on the sub gizmo
	*/
	bool isIdle();
	/**
		returns true iff no editing operation is in process and the mouse cursor is resting on the sub gizmo
	*/
	bool isHovering();
	/**
		returns true iff editing operation is in process
	*/
	bool isEditing();

	/**
		Should be called by the sub class to set the state to idle
	*/
	void setIdle();
	/**
		Should be called by the sub class to set the state to editing
	*/
	void setEditing();

private:
	enum SubGizmoState
	{
		IDLE,
		HOVERING,
		EDITING
	};

	Scene* _scene;

	SceneNode* _target;

	SubGizmoState _state;
};

/**
	Specific gizmo that allows the user to scale an object by dragging the mouse. Represented by a white cube.
*/
class ScaleGizmo : public SubGizmo
{
public:
	ScaleGizmo(Scene* scene);

	virtual void render(Shader* shader);
	virtual UserAction press();

	virtual void update();

private:
	Mesh* _mesh;

	int _mouseStartX, _mouseStartY;
	float _scaleStart;

	static GLfloat VERTEX_POSITION[], VERTEX_COLOR[];
	static GLuint VERTEX_INDICES[];
};

/**
	Specific gizmo that allows the user to translate an object along a plane by dragging the mouse. Represented by a colorcoded quad.
	Red = yz-plane, Blue = xz-plane, Green = xy-plane
*/
class TranslateGizmo : public SubGizmo
{
public:
	// x=0, y=1, z=2
	/**
		Creates a new transform gizmo
		@param axis 0=yz-plane, 1=xz-plane, 2=xy-plane
	*/
	TranslateGizmo(Scene* scene, int axis);

	virtual void render(Shader* shader);
	virtual UserAction press();

	virtual void update();

private:
	
	Mesh* _mesh;
	int _axis;
	
	glm::vec3 _color;
	
	float _mouseOffsetX, _mouseOffsetY;
	Plane _translationPlane;

	static GLfloat VERTEX_POSITION_X[], VERTEX_POSITION_Y[], VERTEX_POSITION_Z[], VERTEX_COLOR[];
};


#endif