#include "JumpFlooding.h"
#include "Geometry.h"

void JumpFlooding::init()
{
	if (_initialized)
		return;

	// Screen quad
	Geometry::createScreenQuad(&_screenQuad);

	// Shader
	_jumpFloodingShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/jumpflood.vert");
	_jumpFloodingShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/jumpflood.frag");
	_jumpFloodingShader.compileShaders();
	_jumpFloodingShader.link("outColor");

	_jumpFloodingMeshShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/jumpfloodmeshshader.vert");
	_jumpFloodingMeshShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/jumpfloodmeshshader.frag");
	_jumpFloodingMeshShader.compileShaders();
	_jumpFloodingMeshShader.link("outColor");

	_jumpFloodResultShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/jumpfloodresult.vert");
	_jumpFloodResultShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/jumpfloodresult.frag");
	_jumpFloodResultShader.compileShaders();
	_jumpFloodResultShader.link("outColor");

	_initialized = true;
}

void JumpFlooding::beginMeshWriting(Framebuffer* src, Framebuffer* dst)
{
	glClearColor(0.f,0.f,0.f,0.f);
	glClearDepth(0.0);
	glDepthFunc(GL_GREATER);
	glCullFace(GL_FRONT);
	glEnable(GL_CULL_FACE);

	dst->bind();
	dst->setViewport();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	dst->unbind();

	src->bind();
	src->setViewport();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	_jumpFloodingMeshShader.bind();

	glUniform2f(_jumpFloodingMeshShader.getUniformLocation("imageSize"),(float)src->getWidth(),(float)src->getHeight());
}

void JumpFlooding::endMeshWriting()
{
	_jumpFloodingMeshShader.unbind();

	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glDrawBuffer(GL_BACK);

	glDepthFunc(GL_LESS);
	glClearDepth(1.0);

	glCullFace(GL_BACK);
}

void JumpFlooding::createDistanceTransformBuffers(Framebuffer& fbo1, Framebuffer& fbo2, int width, int height)
{
	if (!_initialized)
		return;

	FramebufferDescriptor jumpFloodDescr(width, height);
	GLint colorFormat = GL_RGBA;
	GLint colorInternalFormat = GL_RGBA16;
	jumpFloodDescr.colorFormat = &colorFormat;
	jumpFloodDescr.colorInternalFormat = &colorInternalFormat;
	jumpFloodDescr.colorType = GL_UNSIGNED_SHORT;
	jumpFloodDescr.textureWrap = GL_CLAMP_TO_BORDER;

	jumpFloodDescr.depthFormat = GL_DEPTH_COMPONENT;
	jumpFloodDescr.depthInternalFormat = GL_DEPTH_COMPONENT32;

	fbo1.init(jumpFloodDescr);
	fbo2.init(jumpFloodDescr);
}

Shader* JumpFlooding::getMeshShader()
{
	return &_jumpFloodingMeshShader;
}

Framebuffer* JumpFlooding::applyDistanceTransform(Framebuffer* src, Framebuffer* dst)
{
	if (!_initialized)
		return nullptr;

	// jump flooding

	Framebuffer* jumpFloodBuffers[2] = {src, dst};
	int width = src->getWidth();
	int height = src->getHeight();

	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_ALWAYS);
	src->setViewport();

	int srcIndex = 0;
	int dstIndex = 1;
	int longestSide = max(width, height);
	int numIterations = (int)ceil(log((float)longestSide) / log(2.f));

	int kx = width / 2;
	int ky = height / 2;

	_jumpFloodingShader.bind();

	glUniform2f(_jumpFloodingShader.getUniformLocation("invImageSize"), 1.f/(float)width, 1.f/(float)height);

	_screenQuad.bindVAO();

	for (int i=0;i<numIterations;i++)
	{
		jumpFloodBuffers[dstIndex]->bind();

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, jumpFloodBuffers[srcIndex]->getColorTextureHandle());
		glUniform1i(_jumpFloodingShader.getUniformLocation("texture0"), 0);

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, jumpFloodBuffers[srcIndex]->getDepthTextureHandle());
		glUniform1i(_jumpFloodingShader.getUniformLocation("depthTexture"), 1);

		glUniform2ui(_jumpFloodingShader.getUniformLocation("k"), kx, ky);

		_screenQuad.draw();
		
		jumpFloodBuffers[dstIndex]->unbind();

		srcIndex = 1-srcIndex;
		dstIndex = 1-dstIndex;

		kx = (int)ceil((float)kx/2.f);
		ky = (int)ceil((float)ky/2.f);
	}

	_jumpFloodingShader.unbind();

	// put result into w-coordinate of color buffer
	_jumpFloodResultShader.bind();
	jumpFloodBuffers[dstIndex]->bind();

	glUniform2f(_jumpFloodResultShader.getUniformLocation("invImageSize"), 1.f/(float)width, 1.f/(float)height);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, jumpFloodBuffers[srcIndex]->getColorTextureHandle());
	glUniform1i(_jumpFloodResultShader.getUniformLocation("texture0"), 0);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, jumpFloodBuffers[srcIndex]->getDepthTextureHandle());
	glUniform1i(_jumpFloodResultShader.getUniformLocation("depthTexture"), 1);

	_screenQuad.draw();
		
	jumpFloodBuffers[dstIndex]->unbind();
	_jumpFloodResultShader.unbind();

	// reset state
	_screenQuad.unbindVAO();

	glDepthFunc(GL_LESS);

	return jumpFloodBuffers[dstIndex];
}

Mesh JumpFlooding::_screenQuad;
Shader JumpFlooding::_jumpFloodingShader("Jump flooding Shader");
Shader JumpFlooding::_jumpFloodingMeshShader("Jump flooding Mesh Shader");
Shader JumpFlooding::_jumpFloodResultShader("Jump flooding Result Shader");
bool JumpFlooding::_initialized = false;