#include "Mesh.h"
#include "MeshUtils.h"
#include <iostream>
#include <limits>

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif 

using namespace std;

MeshDescriptor::MeshDescriptor()
	: numVertices(-1), vertices(nullptr), normals(nullptr), texCoords(nullptr), colors(nullptr), numIndices(-1),
	indices(nullptr), primitiveMode(GL_TRIANGLES), calcAdjacency(false), hasMinMax(false),
	minX(-1.f), minY(-1.f), minZ(-1.f), maxX(1.f), maxY(1.f), maxZ(1.f)
{
}

Mesh::Mesh()
	: _vao(-1), _numVertices(-1), _vertices(nullptr), _normals(nullptr), _colors(nullptr), _texCoords(nullptr),
	_numIndices(-1), _indices(nullptr), _vboIndices(-1), _texture(nullptr),
	_primitiveMode(GL_TRIANGLES), _initted(false), _hasMinMax(false),
	_minX(-1.f), _minY(-1.f), _minZ(-1.f), _maxX(1.f), _maxY(1.f), _maxZ(1.f)
{
}

Mesh::~Mesh()
{
	if (_initted)
	{
		delete[] _vertices;
		if (hasNormals())
		{
			delete[] _normals;
		}
		if (hasTexCoords())
		{
			delete[] _texCoords;
		}
		if (hasColors())
		{
			delete[] _colors;
		}

		glDeleteBuffers(3, _vboBasic);

		if (_numIndices != -1)
		{
			delete[] _indices;
			glDeleteBuffers(1, &_vboIndices);
		}

		glDeleteVertexArrays(1,&_vao);
	}
}

void Mesh::init(const MeshDescriptor& descr)
{
	// TESTING ///////////////
	/*Vec3 a(vertices[0],vertices[1],vertices[2]);
	Vec3 b(vertices[3],vertices[4],vertices[5]);
	Vec3 c(vertices[6],vertices[7],vertices[8]);

	Vec3 n1 = (b-a).cross(c-a);
	Vec3 n2 = (c-b).cross(a-b);
	Vec3 n3 = (a-c).cross(b-c);

	cout << "n1: " << n1 << endl << "n2: " << n2 << "n3: " << n3 << endl;*/
	//////////////////////////

	_hasMinMax = descr.hasMinMax;
	if (_hasMinMax)
	{
		_minX = descr.minX;
		_minY = descr.minY;
		_minZ = descr.minZ;
		_maxX = descr.maxX;
		_maxY = descr.maxY;
		_maxZ = descr.maxZ;
	}

	bool haveToCalcAdjacency = descr.calcAdjacency;

	if (haveToCalcAdjacency && (descr.normals!=nullptr || descr.texCoords!=nullptr || descr.colors!=nullptr))
	{
		cout << "WARNING: Adjacency only supported for meshes without normals, colors or texture coordinates. Turning off adjacency!" << endl;
		haveToCalcAdjacency = false;
	}

	if (descr.primitiveMode == GL_TRIANGLES || descr.primitiveMode == GL_TRIANGLES_ADJACENCY)
	{
		_primitiveMode = haveToCalcAdjacency ? GL_TRIANGLES_ADJACENCY : GL_TRIANGLES;
	} else if (descr.primitiveMode == GL_TRIANGLE_STRIP || descr.primitiveMode == GL_TRIANGLE_STRIP_ADJACENCY)
	{
		_primitiveMode = haveToCalcAdjacency ? GL_TRIANGLE_STRIP_ADJACENCY : GL_TRIANGLE_STRIP;
	} else if (descr.primitiveMode == GL_LINES || descr.primitiveMode == GL_LINES_ADJACENCY)
	{
		_primitiveMode = haveToCalcAdjacency ? GL_LINES_ADJACENCY : GL_LINES;
	} else if (descr.primitiveMode == GL_LINE_STRIP || descr.primitiveMode == GL_LINE_STRIP_ADJACENCY)
	{
		_primitiveMode = haveToCalcAdjacency ? GL_LINE_STRIP_ADJACENCY : GL_LINE_STRIP;
	} else
	{
		_primitiveMode = descr.primitiveMode;
	}

	
	// Vertex indices
	bool haveToSetVertices = true;
	if (descr.numIndices > 0)
	{
		if (haveToCalcAdjacency)
		{
			calcAdjacency(descr.numIndices, descr.indices, descr.numVertices, descr.vertices, _primitiveMode);
			haveToSetVertices = false;
		}
		else
		{
			_numIndices = descr.numIndices;
			_indices = new GLuint[descr.numIndices];
			memcpy(_indices, descr.indices, sizeof(GLuint) * descr.numIndices);
		}
	}

	if (haveToSetVertices)
	{
		_numVertices = descr.numVertices;
		_vertices = new GLfloat[_numVertices*3];
		memcpy(_vertices, descr.vertices, sizeof(GLfloat) * 3 * _numVertices);
	}

	if (descr.normals == nullptr)
	{
		_normals = nullptr;
	}
	else
	{
		_normals = new GLfloat[descr.numVertices*3];
		memcpy(_normals, descr.normals, sizeof(GLfloat) * 3 * descr.numVertices);
	}

	if (descr.texCoords == nullptr)
	{
		_texCoords = nullptr;
	}
	else
	{
		_texCoords = new GLfloat[descr.numVertices*2];
		memcpy(_texCoords, descr.texCoords, sizeof(GLfloat) * 2 * descr.numVertices);
	}

	if (descr.colors == nullptr)
	{
		_colors = nullptr;
	}
	else
	{
		_colors = new GLfloat[descr.numVertices*3];
		memcpy(_colors, descr.colors, sizeof(GLfloat) * 3 * descr.numVertices);
	}

	glGenVertexArrays(1, &_vao);
	glBindVertexArray(_vao);

	int numBuffers = 1 + (descr.normals == nullptr ? 0 : 1) + (descr.texCoords == nullptr ? 0 : 1)
		+ (descr.colors == nullptr ? 0 : 1);

	glGenBuffers(numBuffers, _vboBasic);

	int bufferIndex = 0;
	
	// Vertex positions
	glBindBuffer(GL_ARRAY_BUFFER, _vboBasic[bufferIndex++]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 3 * _numVertices, _vertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, NULL);

	// Normals
	if (_normals != nullptr)
	{
		glBindBuffer(GL_ARRAY_BUFFER, _vboBasic[bufferIndex++]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 3 * _numVertices, _normals, GL_STATIC_DRAW);
		glEnableVertexAttribArray(2);
		glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 0, NULL);
	}

	// Texture coordinates
	if (_texCoords != nullptr)
	{
		glBindBuffer(GL_ARRAY_BUFFER, _vboBasic[bufferIndex++]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 2 * _numVertices, _texCoords, GL_STATIC_DRAW);
		glEnableVertexAttribArray(3);
		glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, 0, NULL);
	}

	// Colors
	if (_colors != nullptr)
	{
		glBindBuffer(GL_ARRAY_BUFFER, _vboBasic[bufferIndex++]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 3 * _numVertices, _colors, GL_STATIC_DRAW);
		glEnableVertexAttribArray(4);
		glVertexAttribPointer(4, 3, GL_FLOAT, GL_FALSE, 0, NULL);
	}

	if (_numIndices > 0)
	{
		glGenBuffers(1, &_vboIndices);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _vboIndices);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLuint) * _numIndices, _indices, GL_STATIC_DRAW);
	}

	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	_initted = true;
}

void Mesh::init(int numVertices, GLfloat* vertices, GLfloat* normals, GLfloat* texCoords, GLenum primitiveMode, bool adjacency)
{
	MeshDescriptor descr;
	descr.numVertices = numVertices;
	descr.vertices = vertices;
	descr.normals = normals;
	descr.texCoords = texCoords;
	descr.primitiveMode = primitiveMode;
	descr.calcAdjacency = adjacency;

	init(descr);
}

void Mesh::init(int numVertices, GLfloat* vertices, GLfloat* normals, GLfloat* texCoords, int numIndices, GLuint* indices, GLenum primitiveMode, bool adjacency)
{
	MeshDescriptor descr;
	descr.numVertices = numVertices;
	descr.vertices = vertices;
	descr.normals = normals;
	descr.texCoords = texCoords;
	descr.numIndices = numIndices;
	descr.indices = indices;
	descr.primitiveMode = primitiveMode;
	descr.calcAdjacency = adjacency;

	init(descr);
}

void Mesh::calcAdjacency(int numIndices, GLuint* indices, int numVertices, GLfloat* vertices, GLenum primitiveMode)
{
	if (primitiveMode != GL_TRIANGLES_ADJACENCY)
	{
		std::cout << "WARNING: Primitive Mode not supported!" << std::endl;

		_numIndices = numIndices;
		_indices = new GLuint[numIndices];
		memcpy(_indices, indices, sizeof(GLuint) * numIndices);

		_numVertices = numVertices;
		_vertices = new GLfloat[numVertices*3];
		memcpy(_vertices, vertices, sizeof(GLfloat) * 3 * numVertices);

		return;
	}

	_numIndices = numIndices * 2;
	_indices = new GLuint[_numIndices];

	MeshUtils::calculateTrianglesAdjacency(numIndices / 3, indices, _indices, numVertices, vertices, _numVertices, &_vertices);
}

void Mesh::draw() const
{
	if (hasIndices())
	{
		glDrawElements(getPrimitiveMode(), getNumIndices(), GL_UNSIGNED_INT, NULL);
	}
	else
	{
		glDrawArrays(getPrimitiveMode(), 0, getNumVertices());
	}
}

void Mesh::calculateMinMax(bool overrideCurrent)
{
	if (_hasMinMax && !overrideCurrent)
		return;

	float maxFloat = numeric_limits<float>::max();

	float minX=maxFloat, minY=maxFloat, minZ=maxFloat;
	float maxX=-maxFloat, maxY=-maxFloat, maxZ=-maxFloat;

	float curPos;
	for (int i=0;i<3*_numVertices;i++)
	{
		curPos = _vertices[i];
		if (curPos > maxX)
			maxX = curPos;
		if (curPos < minX)
			minX = curPos;

		curPos = _vertices[++i];
		if (curPos > maxY)
			maxY = curPos;
		if (curPos < minY)
			minY = curPos;

		curPos = _vertices[++i];
		if (curPos > maxZ)
			maxZ = curPos;
		if (curPos < minZ)
			minZ = curPos;
	}

	_minX = minX;
	_minY = minY;
	_minZ = minZ;
	_maxX = maxX;
	_maxY = maxY;
	_maxZ = maxZ;
}

bool Mesh::hasMinMax() const
{
	return _hasMinMax;
}

void Mesh::getMin(float& outX, float& outY, float& outZ) const
{
	outX = _minX;
	outY = _minY;
	outZ = _minZ;
}

void Mesh::getMax(float& outX, float& outY, float& outZ) const
{
	outX = _maxX;
	outY = _maxY;
	outZ = _maxZ;
}

bool Mesh::hasIndices() const
{
	return _numIndices != -1;
}

bool Mesh::hasNormals() const
{
	return _normals != nullptr;
}

bool Mesh::hasTexCoords() const
{
	return _texCoords != nullptr;
}

bool Mesh::hasColors() const
{
	return _colors != nullptr;
}

int Mesh::getNumIndices() const
{
	return _numIndices;
}

int Mesh::getNumVertices() const
{
	return _numVertices;
}

void Mesh::setTexture(Texture* texture)
{
	_texture = texture;
}

const Texture* Mesh::getTexture() const
{
	return _texture;
}

void Mesh::bindVAO() const
{
	glBindVertexArray(_vao);
}

void Mesh::unbindVAO() const
{
	glBindVertexArray(0);
}

GLenum Mesh::getPrimitiveMode() const
{
	return _primitiveMode;
}

const GLfloat* Mesh::getVertices() const
{
	return _vertices;
}

const GLfloat* Mesh::getNormals() const
{
	return _normals;
}

const GLfloat* Mesh::getTexCoords() const
{
	return _texCoords;
}

const GLfloat* Mesh::getColors() const
{
	return _colors;
}

const GLuint* Mesh::getIndices() const
{
	return _indices;
}