#ifndef _MESH_H_
#define _MESH_H_

#include "include_opengl.h"
#include "Texture.h"

/**
	Encapsulates information needed to build a mesh.
	@see Mesh
*/
class MeshDescriptor
{
public:
	MeshDescriptor();

	int numVertices;
	GLfloat* vertices;
	GLfloat* normals;
	GLfloat* texCoords;
	GLfloat* colors;
	int numIndices;
	GLuint* indices;
	GLenum primitiveMode;
	bool calcAdjacency;

	bool hasMinMax;
	float maxX, maxY, maxZ;
	float minX, minY, minZ;
};

/**
	Encapsulated GL state of a mesh. Consists of a VAO that stores some or all of the following: vertex positions, normals, texture coordinates, vertex colors and indices.
*/
class Mesh
{
public:
	Mesh();
	~Mesh();

	void init(int numVertices, GLfloat* vertices, GLfloat* normals, GLfloat* texCoords,
		GLenum primitiveMode, bool adjacency=false);
	void init(int numVertices, GLfloat* vertices, GLfloat* normals, GLfloat* texCoords, int numIndices, GLuint* indices,
		GLenum primitiveMode, bool adjacency=false);

	/**
		Generates a VAO that holds the data from `descr`. Accords to the layout location conventions used in all the shaders throughout this project.
	*/
	void init(const MeshDescriptor& descr);
	
	/**
		unused
	*/
	void setTexture(Texture* texture); //nullptr allowed
	/**
		unused
	*/
	const Texture* getTexture() const;

	/**
		Binds the VAO that describes this mesh to the GL
	*/
	void bindVAO() const;
	/**
		Unbinds the VAO that describes this mesh to the GL
	*/
	void unbindVAO() const;

	/**
		Issues a draw call to the GL that renders a triangle mesh.
	*/
	void draw() const;
	/**
		unused
	*/
	void drawLines() const;

	bool hasIndices() const;
	bool hasNormals() const;
	bool hasTexCoords() const;
	bool hasColors() const;
	int getNumIndices() const;
	int getNumVertices() const;

	GLenum getPrimitiveMode() const;

	const GLfloat* getVertices() const;
	const GLfloat* getNormals() const;
	const GLfloat* getTexCoords() const;
	const GLfloat* getColors() const;
	const GLuint* getIndices() const;

	/**
		Calculates the minimal axis-aligned bounding box for this mesh.
		@param overrideCurrent if true, the method calculates the AABB anew even if it is already set
	*/
	void calculateMinMax(bool overrideCurrent=false);

	bool hasMinMax() const;
	void getMin(float& outX, float& outY, float& outZ) const;
	void getMax(float& outX, float& outY, float& outZ) const;

private:
	// every mesh has vertex positions, normals and texture coordinates
	GLuint _vao;

	int _numVertices;
	GLfloat* _vertices;
	GLfloat* _normals;
	GLfloat* _texCoords;
	GLfloat* _colors;
	GLuint _vboBasic[3];

	// can have vertex indexing
	int _numIndices; // -1 if unused
	GLuint* _indices; //don't use if nullptr
	GLuint _vboIndices;

	// bounding box
	bool _hasMinMax;
	float _minX, _minY, _minZ;
	float _maxX, _maxY, _maxZ;

	// can have texture
	Texture* _texture; //can be nullptr

	GLenum _primitiveMode; //GL_TRIANGLES, GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN

	bool _initted;

	void calcAdjacency(int numIndices, GLuint* indices, int numVertices, GLfloat* vertices, GLenum primitiveMode);
};


#endif