#include "MeshLoader.h"

#include <iostream>
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>

#include <stack>
#include <vector>
#include <limits>

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif 

using namespace std;

MeshLoadDescription::MeshLoadDescription()
	: loadNormals(false), loadTexCoords(false), loadColors(false), loadTextures(false), moveOrigin(false),
	originX(Unchanged), originY(Unchanged), originZ(Unchanged)
{
}

void MeshLoader::loadCollada(const std::string& path, Mesh* mesh, bool loadNormals, bool loadTexCoords, bool loadColors, bool loadTextures)
{
	MeshLoadDescription descr;
	descr.loadNormals = loadNormals;
	descr.loadTexCoords = loadTexCoords;
	descr.loadColors = loadColors;
	descr.loadTextures = loadTextures;

	loadCollada(path, mesh, descr);
}

void MeshLoader::loadCollada(const std::string& path, Mesh* mesh, const MeshLoadDescription& descr)
{
	Assimp::Importer importer;

	int flags = aiProcess_Triangulate |
		aiProcess_JoinIdenticalVertices |
		aiProcess_SortByPType |
		aiProcess_RemoveRedundantMaterials;

	if (descr.loadNormals)
	{
		flags |= aiProcess_GenNormals;
		//flags |= aiProcess_FixInfacingNormals;
	}

	const aiScene* scene = importer.ReadFile( path, flags);

	/*
	for(unsigned int i=0;i<scene->mNumMaterials;i++)
	{
		cout << "\nMaterial " << i << ": " << endl;
		aiMaterial* mat = scene->mMaterials[i];
		printMaterial(*mat);
	}
	*/
  
	if(!scene)
	{
		cout << "Error while loading Collada file: " << endl << importer.GetErrorString() << endl;
		return;
	}

	stack<aiNode*> nodes;
	stack<aiMatrix4x4> matrices;

	vector<float>* vertices = new vector<float>();
	vector<float>* normals = new vector<float>();
	vector<float>* colors = new vector<float>();
	vector<unsigned int>* indices = new vector<unsigned int>();

	float maxFloat = std::numeric_limits<float>::max();
	aiVector3D minVertex(maxFloat, maxFloat, maxFloat);
	aiVector3D maxVertex(-maxFloat, -maxFloat, -maxFloat);

	int totalVertices = 0;
	int totalFaces = 0;

	nodes.push(scene->mRootNode);
	matrices.push(scene->mRootNode->mTransformation);

	while (!nodes.empty())
	{
		aiNode* elem = nodes.top();
		aiMatrix4x4 matrix = matrices.top();
		aiMatrix4x4 normalMatrix = matrix;
		normalMatrix.Inverse();
		normalMatrix.Transpose();

		nodes.pop();
		matrices.pop();

		for(int i=0; i<(int)elem->mNumMeshes; i++)
		{
			aiMesh* mesh = scene->mMeshes[elem->mMeshes[i]];

			for (int v=0; v<(int)mesh->mNumVertices; v++)
			{
				const aiVector3D& vertex = mesh->mVertices[v];
				aiVector3D tvertex = matrix * vertex;

				if (tvertex.x > maxVertex.x)
					maxVertex.x = tvertex.x;
				if (tvertex.y > maxVertex.y)
					maxVertex.y = tvertex.y;
				if (tvertex.z > maxVertex.z)
					maxVertex.z = tvertex.z;

				if (tvertex.x < minVertex.x)
					minVertex.x = tvertex.x;
				if (tvertex.y < minVertex.y)
					minVertex.y = tvertex.y;
				if (tvertex.z < minVertex.z)
					minVertex.z = tvertex.z;

				vertices->push_back(tvertex.x);
				vertices->push_back(tvertex.y);
				vertices->push_back(tvertex.z);
			}

			if (descr.loadNormals)
			{
				for (int v=0; v<(int)mesh->mNumVertices; v++)
				{
					const aiVector3D& normal = mesh->mNormals[v];
					aiVector3D tnormal = normalMatrix * normal;

					normals->push_back(tnormal.x);
					normals->push_back(tnormal.y);
					normals->push_back(tnormal.z);
				}
			}

			if (descr.loadColors)
			{
				if (mesh->HasVertexColors(0))
				{
					aiColor4D* color = mesh->mColors[0];
					for (int v=0; v<(int)mesh->mNumVertices; v++)
					{
						colors->push_back(color[v].r);
						colors->push_back(color[v].g);
						colors->push_back(color[v].b);
					}
				}
				else if (mesh->mMaterialIndex >= 0)
				{
					aiColor4D col;
					scene->mMaterials[mesh->mMaterialIndex]->Get<aiColor4D>(AI_MATKEY_COLOR_DIFFUSE, col);
					for (int v=0; v<(int)mesh->mNumVertices; v++)
					{
						colors->push_back(col.r);
						colors->push_back(col.g);
						colors->push_back(col.b);
					}
				}
				else
				{
					for (int v=0; v<(int)mesh->mNumVertices; v++)
					{
						colors->push_back(1.0f);
						colors->push_back(1.0f);
						colors->push_back(1.0f);
					}
				}
			}

			unsigned int indexOffset = totalVertices;

			for (unsigned int f=0; f<mesh->mNumFaces; f++)
			{
				const aiFace& face = mesh->mFaces[f];
				
				indices->push_back(face.mIndices[0] + indexOffset);
				indices->push_back(face.mIndices[1] + indexOffset);
				indices->push_back(face.mIndices[2] + indexOffset);
			}

			totalVertices += mesh->mNumVertices;
			totalFaces += mesh->mNumFaces;
		}
		// END evaluate node

		for (int i=(int)elem->mNumChildren-1; i>=0; i--)
		{
			nodes.push(elem->mChildren[i]);
			matrices.push(matrix * elem->mChildren[i]->mTransformation);
		}
	}

	GLfloat* normalPointer = nullptr;
	if (descr.loadNormals)
	{
		normalPointer = &((*normals)[0]);
	}

	GLfloat* colorPointer = nullptr;
	if (descr.loadColors)
	{
		colorPointer = &((*colors)[0]);
	}

	if(descr.moveOrigin && (descr.originX != Unchanged || descr.originY != Unchanged || descr.originZ != Unchanged))
	{
		float move[3] = {0.0f,0.0f,0.0f};
		Origin origin[3] = {descr.originX, descr.originY, descr.originZ};

		for (int i=0; i<3; i++)
		{
			switch(origin[i])
			{
			case Min:
				move[i] = -minVertex[i];
				break;
			case Max:
				move[i] = -maxVertex[i];
				break;
			case Center:
				move[i] = -(maxVertex[i] + minVertex[i]) / 2.f;
				break;
			}
		}

		for (int i=0; i<totalVertices; i++)
		{
			(*vertices)[3*i  ] += move[0];
			(*vertices)[3*i+1] += move[1];
			(*vertices)[3*i+2] += move[2];
		}

		minVertex.x += move[0];
		minVertex.y += move[1];
		minVertex.z += move[2];

		maxVertex.x += move[0];
		maxVertex.y += move[1];
		maxVertex.z += move[2];
	}

	MeshDescriptor outDescr;
	outDescr.numVertices = totalVertices;
	outDescr.vertices = &((*vertices)[0]);
	outDescr.normals = normalPointer;
	outDescr.colors = colorPointer;
	outDescr.texCoords = nullptr;
	outDescr.numIndices = totalFaces * 3;
	outDescr.indices = &((*indices)[0]);
	outDescr.hasMinMax = true;
	outDescr.minX = minVertex.x;
	outDescr.minY = minVertex.y;
	outDescr.minZ = minVertex.z;
	outDescr.maxX = maxVertex.x;
	outDescr.maxY = maxVertex.y;
	outDescr.maxZ = maxVertex.z;

	cout << "Loaded model from " << path << endl;

	mesh->init(outDescr);
}

void MeshLoader::printMaterial(const aiMaterial& material)
{
	for(unsigned int j=0; j<material.mNumProperties; j++)
	{
		aiMaterialProperty* prop = material.mProperties[j];
		cout << "  " << prop->mKey.C_Str() << " (";
		switch(prop->mType)
		{
		case(aiPTI_Float):
			{cout << "float): ";
			float* p = (float*) prop->mData;
			for(unsigned int k=0; k < (prop->mDataLength/sizeof(float)); k++)
				cout << p[k] << " ";
				
			break;}

		case(aiPTI_String):
			cout << "string): ";
			for(unsigned int k=0; k < prop->mDataLength; k++)
				cout << prop->mData[k];

			break;

		case(aiPTI_Integer):
			{cout << "int): ";
			int* p = (int*) prop->mData;
			for(unsigned int k=0; k < (prop->mDataLength/sizeof(int)); k++)
				cout << p[k] << " ";

			break;}

		case(aiPTI_Buffer):
			cout << "bin)";break;

		default:
			cout << "n/a)";break;

		}
		cout << endl;
	}
}