#include "Scene.h"

Scene::Scene()
	: _globalAngle(DEFAULT_GLOBAL_ANGLE), _edgeCompression(true), _directionalConstraint(true), _ghostDistance(4.f), _lineType(LINES_OFF)
{
	_root = new SceneNode(this);
	_gizmo = new Gizmo(this);
}

SceneNode* Scene::getRoot()
{
	return _root;
}

SceneNode* Scene::createSceneNode()
{
	return new SceneNode(this);
}

void Scene::setCamera(Camera* camera)
{
	_camera = camera;
}

Camera* Scene::getCamera()
{
	return _camera;
}

void Scene::updateTransforms(const glm::mat4& viewMatrix)
{
	_root->propagateTransform(viewMatrix,1.f);
}

void Scene::addRenderable(Renderable* renderable)
{
	_unselectedRenderables.insert(renderable);
}

void Scene::removeRenderable(Renderable* renderable)
{
	_unselectedRenderables.erase(renderable);
	_selectedRenderables.erase(renderable);
}

void Scene::select(Renderable* renderable)
{
	rset_iter_t iter = _unselectedRenderables.find(renderable);
	if (iter != _unselectedRenderables.end())
	{
		_unselectedRenderables.erase(iter);
		_selectedRenderables.insert(renderable);
		renderable->select();
	}
}

Renderable* Scene::getFirstSelected()
{
	if (_selectedRenderables.empty())
		return nullptr;
	return *(_selectedRenderables.begin());
}

rset_iter_t Scene::selectionBegin()
{
	return _selectedRenderables.begin();
}

rset_iter_t Scene::selectionEnd()
{
	return _selectedRenderables.end();
}

void Scene::deselect(Renderable* renderable)
{
	rset_iter_t iter = _selectedRenderables.find(renderable);
	if (iter != _selectedRenderables.end())
	{
		_selectedRenderables.erase(iter);
		_unselectedRenderables.insert(renderable);
		renderable->deselect();
	}
}

void Scene::deselectAll()
{
	for(rset_iter_t iter = _selectedRenderables.begin();
		iter != _selectedRenderables.end();
		iter++)
	{
		(*iter)->deselect();
	}
	_unselectedRenderables.insert(_selectedRenderables.begin(), _selectedRenderables.end());
	_selectedRenderables.clear();
}

void Scene::enableGizmo(bool enabled)
{
	_gizmoEnabled = enabled;
	if (!enabled)
	{
		_gizmo->stopEdit();
	}
}

void Scene::renderAll(Shader* shader)
{
	renderSelected(shader);
	renderUnselected(shader);
}

void Scene::renderSelected(Shader* shader)
{
	for(rset_iter_t iter = _selectedRenderables.begin();
		iter != _selectedRenderables.end();
		iter++)
	{
		(*iter)->render(shader);
	}
}

void Scene::renderSelectedThatCanCut(Shader* shader)
{
	for(rset_iter_t iter = _selectedRenderables.begin();
		iter != _selectedRenderables.end();
		iter++)
	{
		if ((*iter)->canCut())
		{
			(*iter)->render(shader);
		}
	}
}

void Scene::renderUnselected(Shader* shader)
{
	for(rset_iter_t iter = _unselectedRenderables.begin();
		iter != _unselectedRenderables.end();
		iter++)
	{
		(*iter)->render(shader);
	}
}

void Scene::renderUnselectedThatCanBeCut(Shader* shader)
{
	for(rset_iter_t iter = _unselectedRenderables.begin();
		iter != _unselectedRenderables.end();
		iter++)
	{
		if ((*iter)->canBeCut())
		{
			(*iter)->render(shader);
		}
	}
}

void Scene::renderUnselectedThatCannotBeCut(Shader* shader)
{
	for(rset_iter_t iter = _unselectedRenderables.begin();
		iter != _unselectedRenderables.end();
		iter++)
	{
		if (!(*iter)->canBeCut())
		{
			(*iter)->render(shader);
		}
	}
}

void Scene::update(double frameLength)
{
	_gizmo->update();
}

void Scene::renderInterface(Shader* shader)
{
	if (_gizmoEnabled)
	{
		_gizmo->render(shader);
	}
}

void Scene::setSelectionCanCut(bool c)
{
	for(rset_iter_t iter = _selectedRenderables.begin();
		iter != _selectedRenderables.end();
		iter++)
	{
		(*iter)->setCanCut(c);
	}
}

void Scene::setSelectionCanBeCut(bool c)
{
	for(rset_iter_t iter = _selectedRenderables.begin();
		iter != _selectedRenderables.end();
		iter++)
	{
		(*iter)->setCanBeCut(c);
	}
}

void Scene::setSelectionLocalAngle(float angle)
{
	for(rset_iter_t iter = _selectedRenderables.begin();
		iter != _selectedRenderables.end();
		iter++)
	{
		(*iter)->setLocalAngle(angle);
	}
}

Gizmo* Scene::getGizmo()
{
	return _gizmo;
}

void Scene::clear()
{
	deselectAll();
	_root->removeAllSceneNodes();
}

void Scene::setGlobalAngle(float angle)
{
	_globalAngle = std::min(90.f,angle);
}

void Scene::resetGlobalAngle()
{
	_globalAngle = DEFAULT_GLOBAL_ANGLE;
}

float Scene::getGlobalAngle()
{
	return _globalAngle;
}

void Scene::setEdgeCompression(bool c)
{
	_edgeCompression = c;
}

bool Scene::getEdgeCompression()
{
	return _edgeCompression;
}

void Scene::setDirectionalConstraint(bool c)
{
	_directionalConstraint=c;
}

bool Scene::getDirectionalConstraint()
{
	return _directionalConstraint;
}

void Scene::setGhostDistance(float dist)
{
	_ghostDistance = std::max(std::min(4.f, dist), 0.f);
}

float Scene::getGhostDistance()
{
	return _ghostDistance;
}

void Scene::setLineType(int lineType)
{
	_lineType = lineType;
}

int Scene::getLineType()
{
	return _lineType;
}


const float Scene::DEFAULT_GLOBAL_ANGLE = 30.f;

const int Scene::LINES_ON = 0;
const int Scene::LINES_GHOST = 1;
const int Scene::LINES_OFF = 2;