#ifndef _SCENE_H_
#define _SCENE_H_

#include "SceneNode.h"
#include "Renderable.h"
#include "Shader.h"
#include "Gizmo.h"
#include "Camera.h"
#include <unordered_set>

typedef std::unordered_set<Renderable*> rset_t;
typedef std::unordered_set<Renderable*>::iterator rset_iter_t;


/**
	Scene graph. Contains functionality for selecting nodes, setting properties specific to adaptive cutaways.
*/
class Scene
{
public:
	friend class SceneNode;

	Scene();

	/**
		Returns a new scene node of this scene
	*/
	SceneNode* createSceneNode();

	/**
		Returns the root of the scene graph
	*/
	SceneNode* getRoot();

	/**
		Called by the application every frame to update logic of the scene graph
	*/
	void update(double frameLength);
	/**
		Propagtes transformation matrices through the scene graph
	*/
	void updateTransforms(const glm::mat4& viewMatrix);

	/**
		Marks an object as selected
	*/
	void select(Renderable* renderable);
	/**
		Marks an object as not selected
	*/
	void deselect(Renderable* renderable);
	/**
		Deselects all objects
	*/
	void deselectAll();
	/**
		Returns an arbitrary selected object
	*/
	Renderable* getFirstSelected();
	/**
		Returns an iterator to the beginning of the list of selected objects
	*/
	rset_iter_t selectionBegin();
	/**
		Returns an iterator to the end of the list of selected objects
	*/
	rset_iter_t selectionEnd();

	/**
		Draws all renderables in the graph with the selected shader.
	*/
	void renderAll(Shader* shader);
	/**
		Draws all selected renderables in the graph with the selected shader.
	*/
	void renderSelected(Shader* shader);
	/**
		Draws all selected renderables in the graph that have the `canCut` property enabled with the selected shader.
	*/
	void renderSelectedThatCanCut(Shader* shader);
	/**
		Draws all unselected renderables in the graph with the selected shader.
	*/
	void renderUnselected(Shader* shader);
	/**
		Draws all unselected renderables in the graph that have the `canBeCut` property disabled with the selected shader.
	*/
	void renderUnselectedThatCannotBeCut(Shader* shader);
	/**
		Draws all unselected renderables in the graph that have the `canBeCut` property enabled with the selected shader.
	*/
	void renderUnselectedThatCanBeCut(Shader* shader);

	/**
		Set the `canCut` property of every selected renderable to `c`.
	*/
	void setSelectionCanCut(bool c);
	/**
		Set the `canBeCut` property of every selected renderable to `c`.
	*/
	void setSelectionCanBeCut(bool c);
	/**
		Set the local angle property of every selected renderable to `c`.
	*/
	void setSelectionLocalAngle(float angle);

	/**
		Render the gizmo.
	*/
	void renderInterface(Shader* shader);

	/**
		Set the active camera in the scene.
	*/
	void setCamera(Camera* camera);
	/**
		Returns the active camera in the scene.
	*/
	Camera* getCamera();

	/**
		Returns the active gizmo in the scene.
	*/
	Gizmo* getGizmo();

	/**
		Enables or disables [gizmo](@ref Gizmo) editing in the scene.
	*/
	void enableGizmo(bool enabled);

	/**
		Destroys the subgraph of the root and all its attached renderables.
	*/
	void clear();

	/**
		Sets the angle of the cutaway for objects that don't have a local angle set.
	*/
	void setGlobalAngle(float angle);
	/**
		Returns the angle of the cutaway for objects that don't have a local angle set.
	*/
	float getGlobalAngle();
	/**
		Resets the angle of the cutaway for objects that don't have a local angle set to 30.
	*/
	void resetGlobalAngle();
	
	/**
		Enables or disables the edge compression effect.
	*/
	void setEdgeCompression(bool c);
	/**
		Returns true iff the edge compression effect is enabled.
	*/
	bool getEdgeCompression();

	/**
		Enables or disables the directional constraint effect.
	*/
	void setDirectionalConstraint(bool c);
	/**
		Returns true iff the directional constraint effect is enabled.
	*/
	bool getDirectionalConstraint();

	/**
		Sets a value for the ghost lines effect. 0 = lines close to cutaway objects, 4 = maximal distance to cutaway objects
	*/
	void setGhostDistance(float dist);
	/**
		Returns the value for the ghost lines effect.
	*/
	float getGhostDistance();

	/**
		Sets the line type.
		LINES_ON for fully visible outlines of occluding objects.
		LINES_GHOST for ghost lines that are only invisible in front of the cutaway objects. 
		LINES_OFF for no lines on occluding objects.
	*/
	void setLineType(int lineType);
	/**
		Returns the selected line type. See `setLineType`.
	*/
	int getLineType();

	static const int LINES_ON, LINES_GHOST, LINES_OFF;

private:
	SceneNode* _root;
	rset_t _unselectedRenderables;
	rset_t _selectedRenderables;
	Camera* _camera;

	Gizmo* _gizmo;

	bool _gizmoEnabled;
	bool _edgeCompression;
	bool _directionalConstraint;
	float _ghostDistance;
	int _lineType;

	float _globalAngle;

	void addRenderable(Renderable* renderable);
	void removeRenderable(Renderable* renderable);

	static const float DEFAULT_GLOBAL_ANGLE;
	
};


#endif