#include "SceneNode.h"
#include "Renderable.h"
#include "Scene.h"

using namespace std;

SceneNode::SceneNode(Scene* scene)
	: _childNodes(), _parentNode(nullptr), _transformables(), _transform(&_identity), _scene(scene),
	_totalTransform(glm::mat4(1.f)), _totalScale(1.f)
{
}

SceneNode::~SceneNode()
{
	removeAllTransformables();

	for (list<SceneNode*>::iterator iter = _childNodes.begin();
		iter != _childNodes.end();
		iter++)
	{
		delete *iter;
	}
}

void SceneNode::propagateTransform(const glm::mat4& inhTransform, float inhScale)
{
	_transform->update();
	_totalTransform = _transform->propagate(inhTransform, inhScale, &_totalScale);
	
	for (list<Transformable*>::iterator iter = _transformables.begin();
		iter != _transformables.end();
		iter++)
	{
		(*iter)->setTransform(_totalTransform);
	}

	for (list<SceneNode*>::iterator iter = _childNodes.begin();
		iter != _childNodes.end();
		iter++)
	{
		(*iter)->propagateTransform(_totalTransform, _totalScale);
	}
}

void SceneNode::addSceneNode(SceneNode* sceneNode)
{
	if (sceneNode->_parentNode == this)
		return;

	if (sceneNode->_parentNode != nullptr)
	{
		sceneNode->removeFromParent();
	}

	_childNodes.push_back(sceneNode);
	sceneNode->_parentNode = this;
}

void SceneNode::removeSceneNode(SceneNode* sceneNode)
{
	if (sceneNode->_parentNode == this)
	{
		sceneNode->_parentNode = nullptr;
		_childNodes.remove(sceneNode);
		delete sceneNode;
	}
}

void SceneNode::removeAllSceneNodes()
{
	for (list<SceneNode*>::iterator iter = _childNodes.begin();
		iter != _childNodes.end();
		iter++)
	{
		delete *iter;
	}

	_childNodes.clear();
}

void SceneNode::addTransformable(Transformable* transformable)
{
	if (transformable->getParent() == this)
		return;

	if (transformable->getParent() != nullptr)
	{
		transformable->getParent()->removeTransformable(transformable);
	}

	Renderable* renderable = dynamic_cast<Renderable*>(transformable);
	if (renderable)
	{
		_scene->addRenderable(renderable);
	}

	_transformables.push_back(transformable);
	transformable->setParent(this);
}

void SceneNode::removeAllTransformables()
{
	for (list<Transformable*>::iterator iter = _transformables.begin();
		iter != _transformables.end();
		iter++)
	{
		Renderable* renderable = dynamic_cast<Renderable*>(*iter);
		if (renderable)
		{
			_scene->removeRenderable(renderable);
		}

		delete *iter;
	}

	_transformables.clear();
}

void SceneNode::removeTransformable(Transformable* transformable)
{
	if (transformable->getParent() == this)
	{
		transformable->setParent(nullptr);
		_transformables.remove(transformable);

		Renderable* renderable = dynamic_cast<Renderable*>(transformable);
		if (renderable)
		{
			_scene->removeRenderable(renderable);
		}

		delete transformable;
	}
}

void SceneNode::removeFromParent()
{
	_parentNode->removeSceneNode(this);
}

void SceneNode::setTransform(Transform* transform)
{
	_transform = transform;
}

Transform* SceneNode::getTransform()
{
	return (_transform == &_identity ? nullptr : _transform);
}

const glm::mat4& SceneNode::getTotalTransform() const
{
	return _totalTransform;
}

float SceneNode::getTotalScale()
{
	return _totalScale;
}

SceneNode* SceneNode::getParent()
{
	return _parentNode;
}

const std::list<SceneNode*>& SceneNode::getChildNodes() const
{
	return _childNodes;
}

const std::list<Transformable*>& SceneNode::getTransformables() const
{
	return _transformables;
}

IdentityTransform SceneNode::_identity = IdentityTransform();