#ifndef _SCENENODE_H_
#define _SCENENODE_H_

#include "Transform.h"
#include "IdentityTransform.h"
#include "Transformable.h"
#include <list>

class Scene;

/**
	Node in the scene graph. Has a transform used for propagating transformation matrices. Can have scene nodes as children attached.
	Can have [Transformables](@ref Transformable) attached that will inherit the transform from this scene node.
*/
class SceneNode
{
public:
	friend class Scene;

	/**
		Scene node should be constructed by calling the `createSceneNode` method in `Scene`.
	*/
	SceneNode(Scene* scene);
	~SceneNode();
	
	/**
		Add a scene node to the children of this node. `sceneNode` will be removed from its current parent.
	*/
	void addSceneNode(SceneNode* sceneNode);
	/**
		Removes a child node from this node.
	*/
	void removeSceneNode(SceneNode* sceneNode);
	/**
		Removes this node from its current parent.
	*/
	void removeFromParent();
	/**
		Removes all children from this node.
	*/
	void removeAllSceneNodes();

	/**
		Attaches a transformable to this node. The transformable is removed from its current parent.
	*/
	void addTransformable(Transformable* transformable);
	/**
		Removes a transformable from this node.
	*/
	void removeTransformable(Transformable* transformable);
	/**
		Removes all transformables from this node.
	*/
	void removeAllTransformables();

	/**
		Sets the transform used for transformation matrix propagation.
	*/
	void setTransform(Transform* transform);
	/**
		Returns this node's transform.
	*/
	Transform* getTransform();
	/**
		Returns the transformation matrix created by propating the matrix from the root.
	*/
	const glm::mat4& getTotalTransform() const;
	/**
		Returns the scale factor calculated by propagation from the root.
	*/
	float getTotalScale();
	
	/**
		Returns this node's parent node.
	*/
	SceneNode* getParent();

	const std::list<SceneNode*>& getChildNodes() const;
	const std::list<Transformable*>& getTransformables() const;

private:
	Scene* _scene;

	std::list<SceneNode*> _childNodes;
	SceneNode* _parentNode;

	std::list<Transformable*> _transformables;
	Transform* _transform;
	float _totalScale;
	glm::mat4 _totalTransform;

	void propagateTransform(const glm::mat4& inhTransform, float inhScale);

	static IdentityTransform _identity;
};



#endif _SCENENODE_H_