#include "Shader.h"
#include "fileutil.h"

#include <iostream>

using namespace std;

Shader::Shader()
	: _programHandle(-1), _vertexHandle(-1),/* _tessControlHandle(-1), _tessEvalHandle(-1),*/ _geomHandle(-1), _fragHandle(-1),
	_compiled(false), _linked(false), _name("unnamed shader program")
{
}

Shader::Shader(const string& name)
	: _programHandle(-1), _vertexHandle(-1),/* _tessControlHandle(-1), _tessEvalHandle(-1),*/ _geomHandle(-1), _fragHandle(-1),
	_compiled(false), _linked(false), _name(name)
{
}

Shader::~Shader()
{
	if (_compiled)
	{
		if (_vertexHandle != -1)
		{
			glDeleteShader(_vertexHandle);
		}
		if (_geomHandle != -1)
		{
			glDeleteShader(_geomHandle);
		}
		if (_fragHandle != -1)
		{
			glDeleteShader(_fragHandle);
		}

		if (_linked)
		{
			glDeleteProgram(_programHandle);
		}
	}
}

bool Shader::attachShaderSource(int type, const std::string& source)
{
	if (_compiled)
		return false;

	GLuint* handle;
	switch(type)
	{
	case GL_VERTEX_SHADER:
		handle = &_vertexHandle;
		break;
	case GL_GEOMETRY_SHADER:
		handle = &_geomHandle;
		break;
	case GL_FRAGMENT_SHADER:
		handle = &_fragHandle;
		break;
	default:
		return false;
	}

	*handle = glCreateShader(type);
	const char* csource = source.c_str();
	glShaderSource(*handle, 1, &csource, NULL);

	return true;
}

bool Shader::attachShaderFile(int type, const std::string& path)
{
	std::string source;
	bool fileok = fileutil::readFile(path, &source);
	if (fileok)
	{
		return attachShaderSource(type, source);
	}
	return false;
}

bool Shader::compileShaders()
{
	if (_compiled)
		return false;

	GLuint handles[] = {_vertexHandle, _geomHandle, _fragHandle};
	bool ret = true;
	for (size_t i=0; i<3; i++)
	{
		if (handles[i] != -1)
		{
			glCompileShader(handles[i]);
			GLint status;
			glGetShaderiv(handles[i], GL_COMPILE_STATUS, &status);
			if (status != GL_TRUE)
			{
				GLint length;
				glGetShaderiv(handles[i], GL_INFO_LOG_LENGTH, &length);
				char* log = new char[length];
				glGetShaderInfoLog(handles[i], length, NULL, log);
				GLint type;
				glGetShaderiv(handles[i], GL_SHADER_TYPE, &type);
				cout << "ERROR when compiling " << getShaderTypeString(type) << " of " << _name << ":" << std::endl;
				cout << log << endl;
				ret = false;
				delete[] log;
			}
		}
	}
	_compiled = ret;
	return ret;
}

bool Shader::link(const std::string& out)
{
	return link(&out, 1);
}

bool Shader::link(const std::string* out, int numAttachments)
{
	if (!_compiled || _linked)
		return false;


	_programHandle = glCreateProgram();
	if (_vertexHandle != -1)
	{
		glAttachShader(_programHandle, _vertexHandle);
	}
	if (_geomHandle != -1)
	{
		glAttachShader(_programHandle, _geomHandle);
	}
	if (_fragHandle != -1)
	{
		glAttachShader(_programHandle, _fragHandle);
	}

	for (int i=0; i<numAttachments; i++)
	{
		glBindFragDataLocation(_programHandle, i, out[i].c_str());
	}

	glLinkProgram(_programHandle);

	GLint status;
	glGetProgramiv(_programHandle, GL_LINK_STATUS, &status);

	if (status != GL_TRUE)
	{
		GLint length;
		glGetProgramiv(_programHandle, GL_INFO_LOG_LENGTH, &length);
		char* log = new char[length];
		glGetProgramInfoLog(_programHandle, length, NULL, log);
		cout << "ERROR when linking " << _name << ":" << endl;
		cout << log << endl;
		delete[] log;

		_linked = false;
	}
	else
	{
		_linked = true;
	}

	return _linked;
}

void Shader::bind()
{
	glUseProgram(_programHandle);
}
void Shader::unbind()
{
	glUseProgram(0);
}

GLint Shader::getUniformLocation(const std::string& name) const
{
	return glGetUniformLocation(_programHandle, name.c_str());
}

GLint Shader::getAttribLocation(const std::string& name) const
{
	return glGetAttribLocation(_programHandle, name.c_str());
}

string Shader::getShaderTypeString(GLint type)
{
	switch (type)
	{
	case GL_VERTEX_SHADER:
		return "vertex shader";
		break;
	case GL_GEOMETRY_SHADER:
		return "geometry shader";
		break;
	case GL_FRAGMENT_SHADER:
		return "fragment shader";
		break;
	default:
		return "shader";
	}
}

const string& Shader::getName() const
{
	return _name;
}