#ifndef _SHADER_H_
#define _SHADER_H_

#include "include_opengl.h"

#include <string>

/**
	Provides functionality to load GLSL source code from a file, compile it, link it and bind it to the context.
*/
class Shader
{
public:
	Shader();
	/**
		Creates a shader with an associated name. Name will be displayed to link error message to it.
	*/
	Shader(const std::string& name);
	~Shader();

	/**
		Attaches source code from a string to this shader. Shader type is one of GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER.
	*/
	bool attachShaderSource(int type, const std::string& source);
	/**
		Attaches source code from a file to this shader. Shader type is one of GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER.
		@param path file relative to the folder shaders/
	*/
	bool attachShaderFile(int type, const std::string& path);

	/**
		Attempt to compile all source code attached to this object. Error messages are displayed at the standard output.
	*/
	bool compileShaders();
	/**
		Links all compiled shaders to form a shader program. `out` is the only frag data location.
	*/
	bool link(const std::string& out);
	/**
		Links all compiled shaders to form a shader program.
		`out` is a pointer to the first element in an array of frag data locations. `numAttachments` is the number of frag data locations.
	*/
	bool link(const std::string* out, int numAttachments);

	/**
		Binds the shader to the context.
	*/
	void bind();
	/**
		Unbinds the currently bound shader from the context.
	*/
	void unbind();

	/**
		Returns the uniform location `name` of this shader.
	*/
	GLint getUniformLocation(const std::string& name) const;
	/**
		Returns the generic vertex attribute location `name` of this shader.
	*/
	GLint getAttribLocation(const std::string& name) const;

	const std::string& getName() const;

private:
	std::string _name;
	GLuint _programHandle;
	GLuint _vertexHandle, /*_tessControlHandle, _tessEvalHandle,*/ _geomHandle, _fragHandle;
	bool _compiled, _linked;

	std::string getShaderTypeString(GLint type);
};

#endif