#include "Terminal.h"
#include "StringUtil.h"
#include <sstream>
#include <glm/gtc/type_ptr.hpp>

void GLFWCALL globalCharacterCallback(int character, int action)
{
	globalTerminal->characterCallback(character, action);
}

void GLFWCALL globalKeyCallback(int key, int action)
{
	globalTerminal->keyCallback(key,action);
}

Terminal::Terminal(Scene* scene, InputHandler* inputHandler, FontRenderer* fontRenderer, Commander* commander)
	: _active(false), _colorShader("Color Shader"), _fontRenderer(fontRenderer), _blinkTimer(0.0), _currentText(), _scene(scene),
	_inputHandler(inputHandler), _requestInactive(true), _commander(commander)
{
	_colorShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/colorshader.vert");
	_colorShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/colorshader.frag");
	_colorShader.compileShaders();
	_colorShader.link("outColor");

	GLfloat quadV[] = {
		-1.f,  1.f, .05f,
		-1.f, -1.f, .05f,
		 0.f,  1.f, .05f,
		 0.f, -1.f, .05f
	};

	_bgMesh = new Mesh();
	_bgMesh->init(4, quadV, nullptr, nullptr, GL_TRIANGLE_STRIP);

	_currentText = new Text();
	_fontRenderer->createText("", "Lucida", scene, _currentText);

	_currentText->setPosition(20,scene->getCamera()->getHeight()-20);
}

Terminal::~Terminal()
{
}

void Terminal::setActive(bool active)
{
	_active = active;
	if (active)
	{
		globalTerminal = this;
		glfwSetCharCallback(globalCharacterCallback);
		glfwSetKeyCallback(globalKeyCallback);
		_requestInactive = false;
	}
	else
	{
		glfwSetCharCallback(NULL);
		glfwSetKeyCallback(NULL);
	}
}

bool Terminal::requestsInactive()
{
	return _requestInactive;
}

bool Terminal::isActive()
{
	return _active;
}

void Terminal::characterCallback(int character, int action)
{
	if (action == GLFW_PRESS)
	{
		_currentLine += character;
	}
}

void Terminal::execute(const std::string& input)
{
	istringstream iss;
	stringstream ss;
	string lower = StringUtil::toLowerCase(input);
	ss << lower;
	
	vector<string> tokens;
	StringUtil::tokenize(input, &tokens);
	
	if (tokens.size()==0)
		return;

	if (tokens[0] == "model" && tokens.size() > 1)
	{
		string fileName = tokens[1];
		Mesh* mesh = _commander->loadModel(fileName);
		_commander->addInstanceToScene(mesh);
	}
	else if (lower == "edit on")
	{
		_scene->enableGizmo(true);
	}
	else if (lower == "edit off")
	{
		_scene->enableGizmo(false);
	}
	else if (tokens[0] == "savescene" && tokens.size() > 1)
	{
		_commander->saveScene(tokens[1]);
	}
	else if (tokens[0] == "loadscene" && tokens.size() > 1)
	{
		_commander->loadScene(tokens[1]);
	}
	else if (lower == "cancut on")
	{
		_scene->setSelectionCanCut(true);
	}
	else if (lower == "cancut off")
	{
		_scene->setSelectionCanCut(false);
	}
	else if (lower == "canbecut on")
	{
		_scene->setSelectionCanBeCut(true);
	}
	else if (lower == "canbecut off")
	{
		_scene->setSelectionCanBeCut(false);
	}
	else if (tokens[0] == "global" || tokens[0] == "globalangle")
	{
		iss = istringstream(tokens[1]);
		float angle;
		iss >> angle;
		_scene->setGlobalAngle(angle);
	}
	else if (tokens[0] == "local" || tokens[0] == "localangle")
	{
		iss = istringstream(tokens[1]);
		float angle;
		iss >> angle;
		_scene->setSelectionLocalAngle(angle);
	}
	else if (lower == "edge on")
	{
		_scene->setEdgeCompression(true);
	}
	else if (lower == "edge off")
	{
		_scene->setEdgeCompression(false);
	}
	else if (tokens[0].substr(0,3) == "dir" && tokens.size()>1)
	{
		if (tokens[1] == "on")
		{
			_scene->setDirectionalConstraint(true);
		}
		else if (tokens[1] == "off")
		{
			_scene->setDirectionalConstraint(false);
		}
	}
	else if ((tokens[0] == "line" || tokens[0] == "lines") && tokens.size()>1)
	{
		if (tokens[1] == "on")
		{
			_scene->setLineType(Scene::LINES_ON);
		}
		else if (tokens[1] == "off")
		{
			_scene->setLineType(Scene::LINES_OFF);
		}
		else if (tokens[1] == "ghost")
		{
			_scene->setLineType(Scene::LINES_GHOST);
		}
	}
	else if (tokens[0] == "ghost" && tokens.size()>1)
	{
		iss = istringstream(tokens[1]);
		float ghost;
		iss >> ghost;
		_scene->setGhostDistance(ghost);
	}
}

void Terminal::keyCallback(int key, int action)
{
	if (action == GLFW_PRESS)
	{
		if (key == GLFW_KEY_BACKSPACE)
		{
			_currentLine = _currentLine.substr(0,_currentLine.length()-1);
		}
		else if (key == GLFW_KEY_TAB)
		{
			_requestInactive = true;
		}
		else if (key == GLFW_KEY_ENTER)
		{
			execute(_currentLine);

			_currentText->setText(_currentLine);
			_previousTexts.push_front(_currentText);

			if (_previousTexts.size() > 20)
			{
				Text* last = _previousTexts.back();
				delete last;
				_previousTexts.pop_back();
			}

			int i=1;
			for (std::deque<Text*>::iterator iter = _previousTexts.begin();
				iter != _previousTexts.end();
				iter++)
			{
				(*iter)->setPosition(20,_scene->getCamera()->getHeight()-20*(++i));
			}

			_currentText = new Text();
			_fontRenderer->createText("", "Lucida", _scene, _currentText);
			_currentLine = "";

			_currentText->setPosition(20,_scene->getCamera()->getHeight()-20);
		}
	}
}

void Terminal::update(double frameLength)
{
	if (!_active)
		return;

	_blinkTimer += frameLength;
	if (_blinkTimer > 0.5)
		_blinkCharacter = "_";
	else
		_blinkCharacter = "";
	if (_blinkTimer > 1.0)
		_blinkTimer -= 1.0;

	_currentText->setText(_currentLine + _blinkCharacter);
}

void Terminal::render()
{
	if (!_active)
		return;

	glEnable(GL_BLEND);
	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	_colorShader.bind();

	glUniform4f(_colorShader.getUniformLocation("color"), 1.f,1.f,1.f,0.6f);

	_bgMesh->bindVAO();
	_bgMesh->draw();
	_bgMesh->unbindVAO();

	_colorShader.unbind();

	_currentText->render();

	for (std::deque<Text*>::iterator iter = _previousTexts.begin();
				iter != _previousTexts.end();
				iter++)
	{
		(*iter)->render();
	}

	glDisable(GL_BLEND);
}