#include "Vis2Window.h"
#include "Shader.h"
#include "Mesh.h"
#include "JumpFlooding.h"
#include "Geometry.h"
#include "MeshLoader.h"
#include "SceneNode.h"
#include "ColorMaterial.h"
#include "MeshInstance.h"
#include "SRTTransform.h"
#include "Scene.h"
#include "Plane.h"
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/glm.hpp>
#include <GL/glfw.h>
#include <iostream>

using namespace std;

void GLFWCALL globalWindowCharacterCallback(int character, int action)
{
	globalWindow->characterCallback(character, action);
}

void GLFWCALL globalWindowKeyCallback(int key, int action)
{
	globalWindow->keyCallback(key, action);
}

Vis2Window::Vis2Window() :
	GLFWGameWindow(800,800,"Visualization 2",false,3,3),
		textureScreenShader("Texture Screen Shader"), depthShader("Depth shader"),
		silhouetteShader("Silhouette Shader"), cellShader("Cell Shader"),
		combineShader("Combine shader"), cutSurfaceShader("Cut surface Shader"), idShader("ID Shader"),
		gizmoShader("Gizmo Shader"), colorShader("Color Shader"), blur1Shader("Blur 1 Shader"), blur2Shader("Blur 2 Shader"),
		width(800), height(800),
		angle(0.f), scale(1.f),
		inputHandler(nullptr),
		_currentAction(IDLE),
		_pressedTab(false), _pressedDel(false),
		_farPlane(400.f)
{
	globalWindow = this;
}

Vis2Window::~Vis2Window()
{
}

void Vis2Window::setPerspective(Shader* shader, const glm::mat4& perspective, float farPlane)
{
	shader->bind();
	glUniformMatrix4fv(shader->getUniformLocation("perspective"), 1, GL_FALSE, glm::value_ptr(perspective));
	glUniform1f(shader->getUniformLocation("farPlane"), farPlane);
	shader->unbind();
}

void Vis2Window::init()
{
	setPrintFPS(true);

	loadShaders();
	inputHandler = new InputHandler();

	glfwSetCharCallback(globalWindowCharacterCallback);
	glfwSetKeyCallback(globalWindowKeyCallback);

	Geometry::createScreenQuad(&screenQuad);

	JumpFlooding::init();
	JumpFlooding::createDistanceTransformBuffers(jumpFloodBuffers[0], jumpFloodBuffers[1], width, height);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glViewport(0,0,width,height);
	glEnable(GL_CULL_FACE);

	// Set light
	_lightPosition = glm::vec3(30.f,20.f,0.0f);

	// Set camera
	camera = new Camera(glm::vec3(0.f,20.f,50.f), glm::vec3(0.f,0.f,0.f), glm::vec3(0.f, 1.f, 0.f),
		width, height, 0.1f, 1000.f, 70.f);
	_cameraController = new CameraController(camera, inputHandler);
	
	// Bind perspective uniform to shaders which need it
	setPerspective(&cellShader, camera->getPerspective(), _farPlane);
	setPerspective(&depthShader, camera->getPerspective(), _farPlane);
	setPerspective(JumpFlooding::getMeshShader(), camera->getPerspective(), _farPlane);
	setPerspective(&cutSurfaceShader, camera->getPerspective(), _farPlane);
	setPerspective(&idShader, camera->getPerspective(), _farPlane);
	setPerspective(&gizmoShader, camera->getPerspective(), _farPlane);
	setPerspective(&colorShader, camera->getPerspective(), _farPlane);

	// Build scene with two couches
	_scene = new Scene();
	_scene->setCamera(camera);

	_commander = new Commander(_scene);
	/*
	Mesh* mesh = _commander->loadModel("helix.dae");

	SceneNode* node = _commander->addInstanceToScene(mesh);
	((SRTTransform*)node->getTransform())->t.x = 20.f;

	node = _commander->addInstanceToScene(mesh);
	((SRTTransform*)node->getTransform())->t.x = -20.f;*/
	_commander->loadScene("s_test.sce");

	// Edge buffer (single channel, 8bit)
	{FramebufferDescriptor edgeDescr(width, height);

	edgeDescr.numColorAttachments = 1;
	int edgeColorFormat = GL_RED;
	int edgeColorInternalFormat = GL_R8;
	edgeDescr.colorFormat = &edgeColorFormat;
	edgeDescr.colorInternalFormat = &edgeColorInternalFormat;

	edgeDescr.depthFormat = GL_DEPTH_COMPONENT;
	edgeDescr.depthInternalFormat = GL_DEPTH_COMPONENT32;

	edgeBuffer.init(edgeDescr);}

	// Cell buffer (color RGBA 8bit, and ID single channel 16 bit)
	{FramebufferDescriptor cellDescr(width, height);

	cellDescr.numColorAttachments = 2;
	int cellColorFormat[] = {GL_RGBA, GL_RED_INTEGER};
	int cellColorInternalFormat[] = {GL_RGBA8, GL_R16UI};
	cellDescr.colorFormat = cellColorFormat;
	cellDescr.colorInternalFormat = cellColorInternalFormat;

	cellDescr.depthFormat = GL_DEPTH_COMPONENT;
	cellDescr.depthInternalFormat = GL_DEPTH_COMPONENT32;

	cellBuffer.init(cellDescr);}

	{FramebufferDescriptor gaussDescr(200.f*(float)width/(float)height, 200.f);
	
	gaussDescr.numColorAttachments = 1;
	int gaussColorFormat = GL_RED;
	int gaussColorInternalFormat = GL_R8;
	gaussDescr.colorFormat = &gaussColorFormat;
	gaussDescr.colorInternalFormat = &gaussColorInternalFormat;

	gaussDescr.colorMagFilter = GL_LINEAR;

	gaussDescr.depthFormat = GL_DEPTH_COMPONENT;
	gaussDescr.depthInternalFormat = GL_DEPTH_COMPONENT16;

	gauss1Buffer.init(gaussDescr);
	gauss2Buffer.init(gaussDescr);}

	//_selection = nullptr;

	_fontRenderer = new FontRenderer();
	_fontRenderer->loadFont("fonts/LucidaSansTypewriter.ttf","Lucida",14);
	_terminal = new Terminal(_scene, inputHandler, _fontRenderer, _commander);
}

void Vis2Window::destroy()
{
}

Framebuffer* Vis2Window::jumpFloodSelection()
{
	Framebuffer* output = nullptr;
	Renderable* selection = _scene->getFirstSelected();
	Transformable* t = dynamic_cast<Transformable*>(selection);
	if (t)
	{
		JumpFlooding::beginMeshWriting(&jumpFloodBuffers[0], &jumpFloodBuffers[1]);

		glm::vec3 view = camera->getPosition()-camera->getTarget();
		view = glm::normalize(view);
		glm::vec3 n(0.0f,1.0f,0.0f);
		float phip = 3.1415f/2.f - acos(glm::dot(view,n));

		float maxtheta = 30.f * 3.1415f/180.f;
		glUniformMatrix4fv(JumpFlooding::getMeshShader()->getUniformLocation("invView"),1,GL_FALSE, glm::value_ptr(camera->getInverseView()));
		glUniform1f(JumpFlooding::getMeshShader()->getUniformLocation("edgeCompression"), _scene->getEdgeCompression()?0.f:1.f);
		glUniform1f(JumpFlooding::getMeshShader()->getUniformLocation("directional"), _scene->getDirectionalConstraint()?1.f:0.f);


		// render back faces that are furthest away

		float globalAngle = _scene->getGlobalAngle();
		for (rset_iter_t iter = _scene->selectionBegin();
			iter != _scene->selectionEnd();
			iter++)
		{
			if ((*iter)->canCut())
			{
				float angle = globalAngle;
				if ((*iter)->getLocalAngle() >= 0.f)
				{
					angle = (*iter)->getLocalAngle();
				}
				glUniform1f(JumpFlooding::getMeshShader()->getUniformLocation("maxTheta"), angle * 3.1415f/180.f);
				(*iter)->render(JumpFlooding::getMeshShader());
			}
		}

		JumpFlooding::endMeshWriting();

		glDisable(GL_CULL_FACE);

		output = JumpFlooding::applyDistanceTransform(&jumpFloodBuffers[0], &jumpFloodBuffers[1]);
		//output = &jumpFloodBuffers[0];

		glEnable(GL_DEPTH_TEST);
	}
	return output;
}

void Vis2Window::renderVisibleFrontSurfaces(Framebuffer* distanceTransform, const glm::vec3& viewLightPosition)
{
	// render cell shaded meshes
	cellBuffer.bind();


	glDrawBuffer(GL_COLOR_ATTACHMENT0);
	glClearColor(1.f,1.f,1.f,1.f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawBuffer(GL_COLOR_ATTACHMENT1);
	glClearColor(0.f,0.f,0.f,1.f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	cellBuffer.bind();

	cellShader.bind();

	// bind depth map if necessary
	glUniform2f(cellShader.getUniformLocation("invImageSize"), 1.f/width, 1.f/height);
	glUniform3f(cellShader.getUniformLocation("viewLightPosition"), viewLightPosition.x, viewLightPosition.y, viewLightPosition.z);
	if (distanceTransform != nullptr)
	{
		glUniform1i(cellShader.getUniformLocation("useDepthMap"), GL_TRUE);
	
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, distanceTransform->getDepthTextureHandle());
		glUniform1i(cellShader.getUniformLocation("depthMap"), 0);
	}
	else
	{
		glUniform1i(cellShader.getUniformLocation("useDepthMap"), GL_FALSE);
	}

	// draw front faces of not selected meshes
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	//_scene->renderUnselected(&cellShader);
	_scene->renderUnselectedThatCanBeCut(&cellShader);

	// draw selected meshes
	glUniform1i(cellShader.getUniformLocation("useDepthMap"), GL_FALSE);

	_scene->renderSelected(&cellShader);
	_scene->renderUnselectedThatCannotBeCut(&cellShader);
	//_scene->renderSelectedThatCanCut(&cellShader);

	cellBuffer.unbind();
	cellShader.unbind();
}

void Vis2Window::renderCutSurfaces(Framebuffer* distanceTransform, const glm::vec3& viewLightPosition)
{
	if (distanceTransform != nullptr)
	{
		glCullFace(GL_FRONT);

		cellBuffer.bind();
		cutSurfaceShader.bind();

		glUniform3f(cutSurfaceShader.getUniformLocation("viewLightPosition"), viewLightPosition.x, viewLightPosition.y, viewLightPosition.z);
		glUniform2f(cutSurfaceShader.getUniformLocation("invImageSize"), 1.f/width, 1.f/height);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, distanceTransform->getDepthTextureHandle());
		glUniform1i(cutSurfaceShader.getUniformLocation("depthMap"), 0);

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, cellBuffer.getDepthTextureHandle());
		glUniform1i(cutSurfaceShader.getUniformLocation("depthFrontFace"), 1);

		_scene->renderUnselected(&cutSurfaceShader);

		cutSurfaceShader.unbind();
		cellBuffer.unbind();
	}

	
}

void Vis2Window::draw()
{
	glClearColor(0.f,0.f,0.f,1.f);

	//glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glCullFace(GL_FRONT);
	

	// Transform matrices
	glm::vec3 viewLightPosition = glm::vec3(camera->getView() * glm::vec4(_lightPosition,1.f));

	// apply jump flooding to selected objects
	Framebuffer* output = jumpFloodSelection();

	renderVisibleFrontSurfaces(output, viewLightPosition);

	renderCutSurfaces(output, viewLightPosition);


	// render all front faces to depth buffer of edgeBuffer
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
	edgeBuffer.bind();

	glClearDepth(1.0);
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	depthShader.bind();

	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	
	_scene->renderAll(&depthShader);

	depthShader.unbind();

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	edgeBuffer.unbind();

	// render and blur cutout objects
	if (_scene->getLineType() == Scene::LINES_GHOST)
	{
		glDisable(GL_DEPTH_TEST);
		glClearColor(0.f,0.f,0.f,1.f);

		// -- render small resolution cutout
		gauss1Buffer.bind();
		gauss1Buffer.setViewport();
		colorShader.bind();

		glClear(GL_COLOR_BUFFER_BIT );
		glUniform4f(colorShader.getUniformLocation("color"),1.f,1.f,1.f,1.f);
		_scene->renderSelectedThatCanCut(&colorShader);

		colorShader.unbind();
		gauss1Buffer.unbind();

		// -- blur horizontally
		gauss2Buffer.bind();
		blur1Shader.bind();

		glUniform1f(blur1Shader.getUniformLocation("multiplier"), _scene->getGhostDistance());
		glUniform2f(blur1Shader.getUniformLocation("invImageSize"), 1.f/gauss2Buffer.getWidth(), 1.f/gauss2Buffer.getHeight());
	
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, gauss1Buffer.getColorTextureHandle());
		glUniform1i(blur1Shader.getUniformLocation("texture0"), 0);

		screenQuad.bindVAO();
		screenQuad.draw();
		screenQuad.unbindVAO();

		blur1Shader.unbind();
		gauss2Buffer.unbind();

		// -- blur vertically
		gauss1Buffer.bind();
		blur2Shader.bind();

		glUniform1f(blur2Shader.getUniformLocation("multiplier"), _scene->getGhostDistance());
		glUniform2f(blur2Shader.getUniformLocation("invImageSize"), 1.f/gauss1Buffer.getWidth(), 1.f/gauss1Buffer.getHeight());
	
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, gauss2Buffer.getColorTextureHandle());
		glUniform1i(blur2Shader.getUniformLocation("texture0"), 0);

		screenQuad.bindVAO();
		screenQuad.draw();
		screenQuad.unbindVAO();

		blur2Shader.unbind();
		gauss1Buffer.unbind();

		glViewport(0,0,getWidth(),getHeight());
	}


	// render edges of full geometry including ghost lines
	glDisable(GL_CULL_FACE);

	edgeBuffer.bind();

	glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);

	silhouetteShader.bind();

	glUniform2f(silhouetteShader.getUniformLocation("invImageSize"), 1.f/(float)width, 1.f/(float)height);
	glUniform2f(silhouetteShader.getUniformLocation("threshold"), 0.0004f, 0.001f);

	if (_scene->getLineType() != Scene::LINES_OFF)
	{
		if (output != nullptr && _scene->getLineType() == Scene::LINES_GHOST)
		{
			glUniform1i(silhouetteShader.getUniformLocation("useDepthMap"), GL_TRUE);

			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, gauss1Buffer.getColorTextureHandle());
			glUniform1i(silhouetteShader.getUniformLocation("depthMap"), 1);
		}
		else
		{
			glUniform1i(silhouetteShader.getUniformLocation("useDepthMap"), GL_FALSE);
		}

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, edgeBuffer.getDepthTextureHandle());
		glUniform1i(silhouetteShader.getUniformLocation("texture0"), 0);

		screenQuad.bindVAO();
		screenQuad.draw();
		screenQuad.unbindVAO();
	}

	// render edges of cut surfaces
	glEnable(GL_BLEND);
	glBlendEquation(GL_MAX);

	glUniform1i(silhouetteShader.getUniformLocation("useDepthMap"), GL_FALSE);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, cellBuffer.getDepthTextureHandle());
	glUniform1i(silhouetteShader.getUniformLocation("texture0"), 0);



	screenQuad.bindVAO();
	screenQuad.draw();
	screenQuad.unbindVAO();

	glDisable(GL_BLEND);

	silhouetteShader.unbind();

	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);

	edgeBuffer.unbind();

	// render IDs of interface into cellBuffer

	cellBuffer.bind();
	GLenum at = GL_COLOR_ATTACHMENT1;
	glDrawBuffers(1, &at);

	glClearDepth(1.0);
	glClear(GL_DEPTH_BUFFER_BIT);

	idShader.bind();
	_scene->renderInterface(&idShader);
	idShader.unbind();
	cellBuffer.unbind();

	// combine cell-shaded meshes and edges on default framebuffer
	glClearColor(1.f,1.f,1.f,1.f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	combineShader.bind();

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, cellBuffer.getColorTextureHandle(0));
	glUniform1i(combineShader.getUniformLocation("texture0"), 0);

	glActiveTexture(GL_TEXTURE0+1);
	glBindTexture(GL_TEXTURE_2D, edgeBuffer.getColorTextureHandle());
	glUniform1i(combineShader.getUniformLocation("texture1"), 1);

	screenQuad.bindVAO();
	screenQuad.draw();
	screenQuad.unbindVAO();

	combineShader.unbind();

	// draw interface
	gizmoShader.bind();
	
	glClear(GL_DEPTH_BUFFER_BIT);

	glDisable(GL_CULL_FACE);

	_scene->renderInterface(&gizmoShader);
	gizmoShader.unbind();

	glClear(GL_DEPTH_BUFFER_BIT);

	_terminal->render();
}

unsigned int Vis2Window::pick(int x, int y)
{
	unsigned int id;
	
	cellBuffer.bind();
	glReadBuffer(GL_COLOR_ATTACHMENT1);
	glReadPixels(x,height-y,1,1,GL_RED_INTEGER,GL_UNSIGNED_INT, &id);
	cellBuffer.unbind();

	return id;
}

unsigned int Vis2Window::pick()
{
	int xPos, yPos;
	glfwGetMousePos(&xPos, &yPos);
	return pick(xPos,yPos);
}

void Vis2Window::characterCallback(int character, int action)
{
}

void Vis2Window::keyCallback(int key, int action)
{
	if (action == GLFW_PRESS)
	{
		if (_currentAction == IDLE)
		{
			if (key == GLFW_KEY_TAB)
			{
				_pressedTab = true;
			}
			else if (key == GLFW_KEY_DEL)
			{
				_pressedDel = true;
			}
		}
	}
}

void Vis2Window::update(double frameLength)
{
	inputHandler->update();
	
	_cameraController->update(frameLength);
	
	unsigned int id = pick();
	Identifiable* ident = Identifiable::byID(id);

	bool terminalActivated = false;

	Renderable* r = _scene->getFirstSelected();
	Transformable* selection = dynamic_cast<Transformable*>(r);

	if (_currentAction == IDLE)
	{
		int mouseWheel = glfwGetMouseWheel();
		if (mouseWheel != 0)
			_cameraController->zoomBy(mouseWheel);

		if (!inputHandler->isMouseDown(GLFW_MOUSE_BUTTON_LEFT) && !inputHandler->isMouseDown(GLFW_MOUSE_BUTTON_RIGHT))
		{
			if (ident != nullptr)
				ident->hover();
		}

		if (inputHandler->isMousePressed(GLFW_MOUSE_BUTTON_RIGHT))
		{
			_cameraController->startOrbit();
			_currentAction = CAMERA_CONTROL;
		}
		else if (inputHandler->isKeyPressed('F') && selection)
		{
			const glm::mat4& mv = selection->getTransform();
			glm::vec4 mvPosition = glm::vec4(mv[3][0],mv[3][1],mv[3][2],1.f);
			glm::vec4 wPosition = _scene->getCamera()->getInverseView() * mvPosition;
			_cameraController->setPivot(glm::vec3(wPosition));
		}
		else if (inputHandler->isKeyPressed('C'))
		{
			_cameraController->setPivot(glm::vec3(0.f));
		}
		else if (_pressedTab)
		{
			_terminal->setActive(true);
			terminalActivated = true;
			_currentAction = TERMINAL;
		}
		else if (_pressedDel)
		{
			if (selection)
			{
				selection->getParent()->removeFromParent();
				deselect();
			}
		}
		else if (inputHandler->isMousePressed(GLFW_MOUSE_BUTTON_LEFT))
		{
			if (ident != nullptr)
			{
				UserAction newAction = ident->press();
				if (newAction != NO_CHANGE)
					_currentAction = newAction;
			}

			if (id == 0)
			{
				deselect();
			}
			else
			{
				MeshInstance* ptr = dynamic_cast<MeshInstance*>(ident);
				if (ptr)
				{
					int multiple = glfwGetKey(GLFW_KEY_LSHIFT) | glfwGetKey(GLFW_KEY_RSHIFT);
					if (!multiple)
						deselect();

					select(ptr);
				}
			}
		}
	}

	if (_currentAction == CAMERA_CONTROL && inputHandler->isMouseReleased(GLFW_MOUSE_BUTTON_RIGHT))
	{
		_cameraController->stopOrbit();
		_currentAction = IDLE;
	}

	if (_currentAction == GIZMO_EDIT && inputHandler->isMouseReleased(GLFW_MOUSE_BUTTON_LEFT))
	{
		_scene->getGizmo()->stopEdit();
		_currentAction = IDLE;
	}

	if (_currentAction == TERMINAL && !terminalActivated)
	{
		_terminal->update(frameLength);
		if (_terminal->requestsInactive())
		{
			_terminal->setActive(false);
			_currentAction = IDLE;
			glfwSetCharCallback(globalWindowCharacterCallback);
			glfwSetKeyCallback(globalWindowKeyCallback);
		}
	}

	glfwSetMouseWheel(0);
	_pressedTab = false;
	_pressedDel = false;

	glm::vec3 toTarget = camera->getTarget()-camera->getPosition();
	float toCenter = sqrt(toTarget.x*toTarget.x+toTarget.z*toTarget.z);
	float up = toTarget.y;
	float angle = atan(up/toCenter);
	angle = angle*180.f/3.1415f;

	_scene->update(frameLength);
	_scene->updateTransforms(camera->getView());
}

void Vis2Window::deselect()
{
	_scene->deselectAll();
	_scene->getGizmo()->setTarget(nullptr);
	_scene->getGizmo()->setVisible(false);
}

void Vis2Window::select(MeshInstance* selection)
{
	_scene->getGizmo()->setTarget(selection->getParent());
	_scene->getGizmo()->setVisible(true);
	
	_scene->select(selection);
}

void Vis2Window::loadShaders()
{
	textureScreenShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/texturescreenshader.vert");
	textureScreenShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/texturescreenshader.frag");
	textureScreenShader.compileShaders();
	textureScreenShader.link("outColor");

	depthShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/depthshader.vert");
	depthShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/depthshader.frag");
	depthShader.compileShaders();
	depthShader.link("outColor");

	silhouetteShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/silhouetteshader.vert");
	silhouetteShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/silhouetteshader.frag");
	silhouetteShader.compileShaders();
	silhouetteShader.link("outColor");

	cellShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/cellshader.vert");
	cellShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/cellshader.frag");
	cellShader.compileShaders();
	string targets[] = { "outColor", "outID" };
	cellShader.link(targets, 2);

	cutSurfaceShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/cutsurfaceshader.vert");
	cutSurfaceShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/cutsurfaceshader.frag");
	cutSurfaceShader.compileShaders();
	cutSurfaceShader.link(targets, 2);
	
	combineShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/combineshader.vert");
	combineShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/combineshader.frag");
	combineShader.compileShaders();
	combineShader.link("outColor");

	gizmoShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/gizmoshader.vert");
	gizmoShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/gizmoshader.frag");
	gizmoShader.compileShaders();
	gizmoShader.link("outColor");

	idShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/idshader.vert");
	idShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/idshader.frag");
	idShader.compileShaders();
	idShader.link("outID");

	colorShader.attachShaderFile(GL_VERTEX_SHADER, "shaders/whiteshader.vert");
	colorShader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/whiteshader.frag");
	colorShader.compileShaders();
	colorShader.link("outColor");

	blur1Shader.attachShaderFile(GL_VERTEX_SHADER, "shaders/blur1shader.vert");
	blur1Shader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/blur1shader.frag");
	blur1Shader.compileShaders();
	blur1Shader.link("outColor");

	blur2Shader.attachShaderFile(GL_VERTEX_SHADER, "shaders/blur2shader.vert");
	blur2Shader.attachShaderFile(GL_FRAGMENT_SHADER, "shaders/blur2shader.frag");
	blur2Shader.compileShaders();
	blur2Shader.link("outColor");
}